
color_names = ('black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white')
foreground = {color_names[x]: '3%s' % x for x in range(8)}
background = {color_names[x]: '4%s' % x for x in range(8)}

RESET = '0'
opt_dict = {'bold': '1', 'underscore': '4', 'blink': '5', 'reverse': '7', 'conceal': '8'}


def colorize(text='', opts=(), **kwargs):
	"""
	Return your text, enclosed in ANSI graphics codes.

	Depends on the keyword arguments 'fg' and 'bg', and the contents of
	the opts tuple/list.

	Return the RESET code if no parameters are given.

	Valid colors:
		'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'

	Valid options:
		'bold'
		'underscore'
		'blink'
		'reverse'
		'conceal'
		'noreset' - string will not be auto-terminated with the RESET code

	Examples:
		colorize('hello', fg='red', bg='blue', opts=('blink',))
		colorize()
		colorize('goodbye', opts=('underscore',))
		print(colorize('first line', fg='red', opts=('noreset',)))
		print('this should be red too')
		print(colorize('and so should this'))
		print('this should not be red')
	"""
	code_list = []
	if text == '' and len(opts) == 1 and opts[0] == 'reset':
		return '\x1b[%sm' % RESET
	for k, v in kwargs.items():
		if k == 'fg':
			code_list.append(foreground[v])
		elif k == 'bg':
			code_list.append(background[v])
	for o in opts:
		if o in opt_dict:
			code_list.append(opt_dict[o])
	if 'noreset' not in opts:
		text = '%s\x1b[%sm' % (text or '', RESET)
	return '%s%s' % (('\x1b[%sm' % ';'.join(code_list)), text or '')


def make_style(opts=(), **kwargs):
	"""
	Return a function with default parameters for colorize()

	Example:
		bold_red = make_style(opts=('bold',), fg='red')
		print(bold_red('hello'))
		KEYWORD = make_style(fg='yellow')
		COMMENT = make_style(fg='blue', opts=('bold',))
	"""
	return lambda text: colorize(text, opts, **kwargs)


NOCOLOR_PALETTE = 'nocolor'
DARK_PALETTE = 'dark'
LIGHT_PALETTE = 'light'

PALETTES = {
	NOCOLOR_PALETTE: {
		'ERROR': {},
		'SUCCESS': {},
		'WARNING': {},
		'NOTICE': {},
		'SQL_FIELD': {},
		'SQL_COLTYPE': {},
		'SQL_KEYWORD': {},
		'SQL_TABLE': {},
		'HTTP_INFO': {},
		'HTTP_SUCCESS': {},
		'HTTP_REDIRECT': {},
		'HTTP_NOT_MODIFIED': {},
		'HTTP_BAD_REQUEST': {},
		'HTTP_NOT_FOUND': {},
		'HTTP_SERVER_ERROR': {},
		'MIGRATE_HEADING': {},
		'MIGRATE_LABEL': {},
	},
	DARK_PALETTE: {
		'ERROR': {'fg': 'red', 'opts': ('bold',)},
		'SUCCESS': {'fg': 'green', 'opts': ('bold',)},
		'WARNING': {'fg': 'yellow', 'opts': ('bold',)},
		'NOTICE': {'fg': 'red'},
		'SQL_FIELD': {'fg': 'green', 'opts': ('bold',)},
		'SQL_COLTYPE': {'fg': 'green'},
		'SQL_KEYWORD': {'fg': 'yellow'},
		'SQL_TABLE': {'opts': ('bold',)},
		'HTTP_INFO': {'opts': ('bold',)},
		'HTTP_SUCCESS': {},
		'HTTP_REDIRECT': {'fg': 'green'},
		'HTTP_NOT_MODIFIED': {'fg': 'cyan'},
		'HTTP_BAD_REQUEST': {'fg': 'red', 'opts': ('bold',)},
		'HTTP_NOT_FOUND': {'fg': 'yellow'},
		'HTTP_SERVER_ERROR': {'fg': 'magenta', 'opts': ('bold',)},
		'MIGRATE_HEADING': {'fg': 'cyan', 'opts': ('bold',)},
		'MIGRATE_LABEL': {'opts': ('bold',)},
	},
	LIGHT_PALETTE: {
		'ERROR': {'fg': 'red', 'opts': ('bold',)},
		'SUCCESS': {'fg': 'green', 'opts': ('bold',)},
		'WARNING': {'fg': 'yellow', 'opts': ('bold',)},
		'NOTICE': {'fg': 'red'},
		'SQL_FIELD': {'fg': 'green', 'opts': ('bold',)},
		'SQL_COLTYPE': {'fg': 'green'},
		'SQL_KEYWORD': {'fg': 'blue'},
		'SQL_TABLE': {'opts': ('bold',)},
		'HTTP_INFO': {'opts': ('bold',)},
		'HTTP_SUCCESS': {},
		'HTTP_REDIRECT': {'fg': 'green', 'opts': ('bold',)},
		'HTTP_NOT_MODIFIED': {'fg': 'green'},
		'HTTP_BAD_REQUEST': {'fg': 'red', 'opts': ('bold',)},
		'HTTP_NOT_FOUND': {'fg': 'red'},
		'HTTP_SERVER_ERROR': {'fg': 'magenta', 'opts': ('bold',)},
		'MIGRATE_HEADING': {'fg': 'cyan', 'opts': ('bold',)},
		'MIGRATE_LABEL': {'opts': ('bold',)},
	}
}
DEFAULT_PALETTE = DARK_PALETTE


def parse_color_setting(config_string):
	"""Parse a DJANGO_COLORS environment variable to produce the system palette

	The general form of a palette definition is:

		"palette;role=fg;role=fg/bg;role=fg,option,option;role=fg/bg,option,option"

	where:
		palette is a named palette; one of 'light', 'dark', or 'nocolor'.
		role is a named style used by Django
		fg is a background color.
		bg is a background color.
		option is a display options.

	Specifying a named palette is the same as manually specifying the individual
	definitions for each role. Any individual definitions following the palette
	definition will augment the base palette definition.

	Valid roles:
		'error', 'success', 'warning', 'notice', 'sql_field', 'sql_coltype',
		'sql_keyword', 'sql_table', 'http_info', 'http_success',
		'http_redirect', 'http_not_modified', 'http_bad_request',
		'http_not_found', 'http_server_error', 'migrate_heading',
		'migrate_label'

	Valid colors:
		'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'

	Valid options:
		'bold', 'underscore', 'blink', 'reverse', 'conceal', 'noreset'
	"""
	if not config_string:
		return PALETTES[DEFAULT_PALETTE]

	# Split the color configuration into parts
	parts = config_string.lower().split(';')
	palette = PALETTES[NOCOLOR_PALETTE].copy()
	for part in parts:
		if part in PALETTES:
			# A default palette has been specified
			palette.update(PALETTES[part])
		elif '=' in part:
			# Process a palette defining string
			definition = {}

			# Break the definition into the role,
			# plus the list of specific instructions.
			# The role must be in upper case
			role, instructions = part.split('=')
			role = role.upper()

			styles = instructions.split(',')
			styles.reverse()

			# The first instruction can contain a slash
			# to break apart fg/bg.
			colors = styles.pop().split('/')
			colors.reverse()
			fg = colors.pop()
			if fg in color_names:
				definition['fg'] = fg
			if colors and colors[-1] in color_names:
				definition['bg'] = colors[-1]

			# All remaining instructions are options
			opts = tuple(s for s in styles if s in opt_dict.keys())
			if opts:
				definition['opts'] = opts

			# The nocolor palette has all available roles.
			# Use that palette as the basis for determining
			# if the role is valid.
			if role in PALETTES[NOCOLOR_PALETTE] and definition:
				palette[role] = definition

	# If there are no colors specified, return the empty palette.
	if palette == PALETTES[NOCOLOR_PALETTE]:
		return None
	return palette



"""
Clock
"""
import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from .views import ClockWidget
from pyplanet.core.signals import pyplanet_start_after


class Clock(AppConfig):
    game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = ClockWidget(self)

    async def on_start(self):
        self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
        self.context.signals.listen(mp_signals.map.map_begin, self.map_start)
        self.context.signals.listen(pyplanet_start_after, self.on_after_start)

    async def player_connect(self, player, **kwargs):
        await self.widget.display(player)

    async def map_start(self, *args, **kwargs):
        await self.widget.display()

    async def on_after_start(self, *args, **kwargs):
        await asyncio.sleep(1)
        asyncio.ensure_future(asyncio.gather(*[
            self.player_connect(p) for p in self.instance.player_manager.online
        ]))



import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.trackmania.callbacks import finish, finish_royal_section


class TrackmaniaConfig(AppConfig):
	name = 'pyplanet.apps.core.trackmania'
	core = True

	app_dependencies = [
		'core.maniaplanet'
	]

	game_dependencies = [
		'trackmania', 'trackmania_next'
	]

	async def on_start(self):
		self.context.signals.register_signal(finish)
		self.context.signals.register_signal(finish_royal_section)

		# TM2020 Checkpoint data enabler.
		if self.instance.game.game == 'tmnext':
			await asyncio.gather(
				self.instance.gbx.script('Trackmania.Event.SetCurRaceCheckpointsMode', 'always', encode_json=False, response_id=False),
				self.instance.gbx.script('Trackmania.Event.SetCurLapCheckpointsMode', 'always', encode_json=False, response_id=False)
			)

			self.instance.signals.listen('maniaplanet:match_begin', self.match_begin_royal)

			if await self.instance.mode_manager.get_current_script() == 'Trackmania/TM_RoyalTimeAttack_Online':
				# Reset game, restart map.
				await self.instance.gbx('RestartMap')

	async def match_begin_royal(self, **kwargs):
		"""
		Handle royal match begin callback.
		"""
		if await self.instance.mode_manager.get_current_script() == 'Trackmania/TM_RoyalTimeAttack_Online':
			for player in self.instance.player_manager.online:
				player.flow.handle_match_begin_royal()



from jinja2 import Environment, select_autoescape

from pyplanet.conf import settings
from pyplanet.core.ui.loader import PyPlanetLoader


async def load_template(file):
	return Template(file)


class _EnvironmentManager:
	def __init__(self):
		self._environment = None

	@property
	def environment(self):
		if not self._environment:
			self._environment = Environment(
				enable_async=True,
				loader=PyPlanetLoader.get_loader(),
				autoescape=select_autoescape(['html', 'xml', 'Txt', 'txt', 'ml', 'ms', 'script.txt', 'Script.Txt']),
				auto_reload=bool(settings.DEBUG),
			)
		return self._environment

EnvironmentManager = _EnvironmentManager()


class Template:
	"""
	Template class manages the template file source and the rendering of it.

	Will also take care of the loader of the Jinja2 template engine.

	Some notable prefixes:

	- core.views: ``pyplanet.views.templates``.
	- core.pyplanet: ``pyplanet.apps.core.pyplanet.templates``.
	- core.maniaplanet: ``pyplanet.apps.core.pyplanet.templates``.
	- core.trackmania: ``pyplanet.apps.core.trackmania.templates``.
	- core.shootmania: ``pyplanet.apps.core.shootmania.templates``.
	- [app_label]: ``[app path]/templates``.
	"""

	def __init__(self, file):
		self.file = file
		self.env = EnvironmentManager.environment
		self.template = self.env.get_template(file)

	async def render(self, **data):
		return await self.template.render_async(**data)



import os
import asynctest

from pyplanet.utils.gbxparser import GbxParser
from tests import TEST_FILES_DIR


class TestGbxParser(asynctest.TestCase):

	@asynctest.fail_on(unused_loop=False)
	async def test_gbxparser_map_file(self):
		parser = GbxParser(file=os.path.join(
			TEST_FILES_DIR,
			'maps',
			'greyroad.gbx'
		))

		map_info = await parser.parse()

		assert map_info['time_bronze'] == 72000
		assert map_info['time_silver'] == 57000
		assert map_info['time_gold'] == 51000
		assert map_info['time_author'] == 47488

		assert map_info['uid'] == '46Yh0hgv5EdSb6IkHsYK1PXHaua'
		assert map_info['name'] == '$s$678$oGrey$o$fff road'
		assert map_info['price'] == 5135
		assert map_info['is_multilap'] == False
		assert map_info['map_type'] == 'Trackmania\\Race'
		assert map_info['map_style'] == ''
		assert map_info['editor'] == 'advanced'
		assert map_info['checkpoints'] == 11
		assert map_info['laps'] == 1

		assert map_info['author_login'] == 'tomvalk'
		assert map_info['author_score'] == 47488
		assert map_info['author_version'] == 0
		assert map_info['author_nickname'] == '$f80$i$s$o$h[maniaflash?toffe]Toffe$z$06fSmurf'
		assert map_info['author_extra'] == ''

		assert map_info['environment'] == 'Canyon'
		assert map_info['title_id'] == 'TMCanyon'
		assert map_info['mood'] == 'Sunrise'

	@asynctest.fail_on(unused_loop=False)
	async def test_gbxparser_map_buffer(self):
		buffer = open(os.path.join(
			TEST_FILES_DIR,
			'maps',
			'greyroad.gbx'
		), mode='rb')
		parser = GbxParser(buffer=buffer)

		map_info = await parser.parse()

		assert map_info['time_bronze'] == 72000
		assert map_info['time_silver'] == 57000
		assert map_info['time_gold'] == 51000
		assert map_info['time_author'] == 47488

		assert map_info['uid'] == '46Yh0hgv5EdSb6IkHsYK1PXHaua'
		assert map_info['name'] == '$s$678$oGrey$o$fff road'
		assert map_info['price'] == 5135
		assert map_info['is_multilap'] == False
		assert map_info['map_type'] == 'Trackmania\\Race'
		assert map_info['map_style'] == ''
		assert map_info['editor'] == 'advanced'
		assert map_info['checkpoints'] == 11
		assert map_info['laps'] == 1

		assert map_info['author_login'] == 'tomvalk'
		assert map_info['author_score'] == 47488
		assert map_info['author_version'] == 0
		assert map_info['author_nickname'] == '$f80$i$s$o$h[maniaflash?toffe]Toffe$z$06fSmurf'
		assert map_info['author_extra'] == ''

		assert map_info['environment'] == 'Canyon'
		assert map_info['title_id'] == 'TMCanyon'
		assert map_info['mood'] == 'Sunrise'



from importlib import import_module

from pyplanet.core.management import CommandError
from pyplanet.core.management.templates import TemplateCommand


class Command(TemplateCommand):
	help = 'Create a new PyPlanet local setup.'

	requires_system_checks = False
	requires_migrations_checks = False

	def handle(self, **options):
		project_name, target = options.pop('name'), options.pop('directory')
		self.validate_name(project_name, 'project')

		# Check that the project_name cannot be imported.
		try:
			if project_name == '.':
				raise ImportError()
			import_module(project_name)
		except ImportError:
			pass
		else:
			raise CommandError(
				"%r conflicts with the name of an existing Python module and "
				"cannot be used as a project name. Please try another name." % project_name
			)

		super().handle('project', project_name, target, **options)



import logging
import os

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.apps.contrib.mx.api import MXApi
from pyplanet.apps.contrib.mx.exceptions import MXMapNotFound, MXInvalidResponse
from pyplanet.apps.contrib.mx.view import MxSearchListView, MxPacksListView, MxStatusListView, MxAwardWidget
from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting
from collections import namedtuple

from pyplanet.utils import gbxparser

logger = logging.getLogger(__name__)


class MX(AppConfig):  # pragma: no cover
	name = 'pyplanet.apps.contrib.mx'
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.api = MXApi()

		self.namespace = 'mx'
		self.site_name = 'ManiaExchange'
		self.site_short_name = 'MX'
		self.award_widget = None

		self.setting_mx_key = Setting(
			'mx_key', 'ManiaExchange/TrackmaniaExchange Key', Setting.CAT_KEYS, type=str, default=None,
			description='Is only required when you want to download from a private group/section!'
		)
		self.setting_display_award_widget = Setting(
			'display_award_widget', 'Display the award widget on podium', Setting.CAT_BEHAVIOUR, type=bool,
			description='Whether to display the widget linking to the MX award page on podium.', default=True
		)

	async def on_init(self):
		# Set server login in the API.
		self.api.server_login = self.instance.game.server_player_login
		# Set the default site in.
		self.api.site = self.instance.game.game
		await self.api.create_session()

	async def on_start(self):
		await self.instance.permission_manager.register(
			'add_remote', 'Add map from remote source (such as MX)', app=self, min_level=2)
		await self.context.setting.register(
			self.setting_mx_key, self.setting_display_award_widget
		)

		self.award_widget = MxAwardWidget(self)

		if self.instance.game.game == 'tmnext':
			self.namespace = 'tmx'
			self.site_name = 'TrackmaniaExchange'
			self.site_short_name = 'TMX'

		await self.instance.command_manager.register(
			Command(command='info', namespace=self.namespace, target=self.mx_info,
					description='Display ManiaExchange/TrackmaniaExchange information for current map.'),
			# support backwards
			Command(command='mx', namespace='add', target=self.add_mx_map, perms='mx:add_remote', admin=True,
					description='Add map from ManiaExchange to the maplist.').add_param(
				'maps', nargs='*', type=str, required=True, help='MX ID(s) of maps to add.'),
			# new mx command random (Adding) Random Maps from MX
			Command(command='random', namespace=self.namespace, target=self.random_mx_map, perms='mx:add_remote',
					admin=True, description='Get Random Maps on ManiaExchange/TrackmaniaExchange.'),
			# new mx namespace
			Command(command='search', aliases=['list'], namespace=self.namespace, target=self.search_mx_map, perms='mx:add_remote',
					admin=True, description='Search for maps on ManiaExchange/TrackmaniaExchange.'),
			Command(command='add', namespace=self.namespace, target=self.add_mx_map, perms='mx:add_remote', admin=True,
					description='Add map from ManiaExchange/TrackmaniaExchange to the maplist.').add_param(
				'maps', nargs='*', type=str, required=True, help='MX/TMX ID(s) of maps to add.'),
			Command(command='status', namespace=self.namespace, target=self.status_mx_maps, perms='mx:add_remote', admin=True,
					description='View the map statuses compared to ManiaExchange/TrackmaniaExchange.'),

			# new mxpack namespace
			Command(command='search', aliases=['list'], namespace='{}pack'.format(self.namespace), target=self.search_mx_pack,
					perms='mx:add_remote', admin=True, description='Search for mappacks on ManiaExchange/TrackmaniaExchange.'),
			Command(command='add', namespace='{}pack'.format(self.namespace), target=self.add_mx_pack, perms='mx:add_remote',
					admin=True, description='Add mappack from ManiaExchange/TrackmaniaExchange to the maplist.')
				.add_param('pack', nargs='*', type=str, required=True, help='MX/TMX ID(s) of mappacks to add.'),
		)

		# Register callback.
		self.context.signals.listen(mp_signals.flow.podium_start, self.podium_start)
		self.context.signals.listen(mp_signals.flow.podium_end, self.podium_end)

	async def podium_start(self, **kwargs):
		if await self.setting_display_award_widget.get_value() is True:
			mx_info = await self.api.map_info(self.instance.map_manager.current_map.uid)
			if mx_info and len(mx_info) >= 1:
				self.award_widget.mx_id = mx_info[0][0]
				await self.award_widget.display()

	async def podium_end(self, **kwargs):
		await self.award_widget.hide()

	async def random_mx_map(self, player, data, **kwargs):
		map_random_id = await self.api.mx_random()
		await self.instance.command_manager.execute(
			player,
			'//{} add maps'.format(self.namespace),
			str(map_random_id)
		)

	async def mx_info(self, player, data, **kwargs):
		try:
			map_info = await self.api.map_info(self.instance.map_manager.current_map.uid)
		except Exception as e:
			map_info = list()
			logger.error('Could not retrieve map info from MX/TM API: {}'.format(str(e)))
		if len(map_info) != 1:
			message = '$f00Map could not be found on MX!'
			await self.instance.chat(message, player)
			return
		map_info = map_info[0][1]

		messages = [
			'$o$ff0{site_name} info:$o Name: $fff{map_name}$ff0, {site_code}-username: $fff{map_username}'.format(
				site_name=self.site_name,
				site_code=self.site_short_name,
				map_name=map_info['Name'],
				map_username=map_info['Username'],
			)
		]
		if 'ReplayCount' in map_info:  # If TM with ReplayCount
			messages.append(
				'$ff0Number of replays: $fff{num_replays}$ff0, Number of awards: $fff{num_awards}$ff0, {site_code}-ID: $l[{link}]$fff{id}$l $n(click to open {site_code})'.format(
					num_replays=map_info['ReplayCount'],
					num_awards=map_info['AwardCount'],
					site_code=self.site_short_name,
					link='{}/s/tr/{}'.format(self.api.base_url(), map_info['TrackID']),
					id=map_info['TrackID'],
				)
			)
		else:
			messages.append(
				'$ff0{site_code}-ID: $l[{link}]$fff{id}$l (click to open {site_code})'.format(
					site_code=self.site_short_name,
					link='{}/s/tr/{}'.format(self.api.base_url(), map_info['MapID']),
					id=map_info['MapID']
				)
			)

		await self.instance.gbx.multicall(*[self.instance.chat(message, player) for message in messages])

	async def search_mx_pack(self, player, data, **kwargs):
		self.api.key = await self.setting_mx_key.get_value()
		window = MxPacksListView(self, player, self.api)
		await window.display()

	async def search_mx_map(self, player, data, **kwargs):
		self.api.key = await self.setting_mx_key.get_value()
		window = MxSearchListView(self, player, self.api)
		await window.display()

	async def status_mx_maps(self, player, data, **kwargs):
		await self.instance.chat('$ff0{site_code}: Please wait, checking for updated maps... This can take a while.'.format(
			site_code=self.site_short_name
		), player)
		self.api.key = await self.setting_mx_key.get_value()
		window = MxStatusListView(self, self.api)
		await window.display(player=player)

	async def add_mx_pack(self, player, data, **kwargs):
		try:
			pack_id = data.pack[0]
			token = data.pack[1] if len(data.pack) == 2 else ""
			ids = await self.api.get_pack_ids(pack_id, token)
			mx_ids = list()
			for obj in ids:
				mx_ids.append(str(obj[0]))

			mock = namedtuple("data", ["maps"])

			await self.instance.chat('$ff0{}: Installing mappack... This can take a while.'.format(self.site_short_name), player)
			await self.add_mx_map(player, mock(maps=mx_ids))
			await self.instance.chat('$ff0{}: Done Installing mappack!'.format(self.site_short_name), player)
		except MXMapNotFound:
			message = '$ff0Error: Can\'t add map pack from {}, due error.'.format(self.site_short_name)
			await self.instance.chat(message, player)

	async def add_mx_map(self, player, data, **kwargs):
		# Make sure we update the key in the api.
		self.api.key = await self.setting_mx_key.get_value()

		# Prepare and fetch information about the maps from MX.
		mx_ids = data.maps

		try:
			infos = await self.api.map_info(*mx_ids)
			if len(infos) == 0:
				raise MXMapNotFound()
		except MXMapNotFound:
			message = '$f00Error: Can\'t add map from {}. Map not found on {}!'.format(self.site_short_name, self.site_name)
			await self.instance.chat(message, player)
			return
		except MXInvalidResponse as e:
			message = '$f00Error: Got invalid response from {}: {}'.format(self.site_name, str(e))
			await self.instance.chat(message, player.login)
			return

		try:
			if not await self.instance.storage.driver.exists(os.path.join('UserData', 'Maps', 'PyPlanet-MX')):
				await self.instance.storage.driver.mkdir(os.path.join('UserData', 'Maps', 'PyPlanet-MX'))
		except Exception as e:
			message = '$f00Error: Can\'t check or create folder: {}'.format(str(e))
			await self.instance.chat(message, player.login)
			return

		# Fetch setting if juke after adding is enabled.
		juke_after_adding = await self.instance.setting_manager.get_setting(
			'admin', 'juke_after_adding', prefetch_values=True)
		juke_maps = await juke_after_adding.get_value()
		if 'jukebox' not in self.instance.apps.apps:
			juke_maps = False
		added_map_uids = list()

		for mx_id, mx_info in infos:
			if 'Name' not in mx_info:
				continue

			try:
				# Test if map isn't yet in our current map list.
				if self.instance.map_manager.playlist_has_map(mx_info['MapUID']):
					raise Exception('Map already in playlist! Update? remove it first!')

				# Download file + save
				resp = await self.api.download(mx_id)
				map_filename = os.path.join('PyPlanet-MX', '{}-{}.Map.Gbx'.format(
					self.instance.game.game.upper(), mx_id
				))
				async with self.instance.storage.open_map(map_filename, 'wb+') as map_file:
					await map_file.write(await resp.read())
					await map_file.close()

				# Insert map to server.
				result = await self.instance.map_manager.add_map(map_filename, save_matchsettings=False)

				if result:
					added_map_uids.append(mx_info['MapUID'])

					message = '$ff0Admin $fff{}$z$s$ff0 has added{} the map $fff{}$z$s$ff0 by $fff{}$z$s$ff0 from {}..'.format(
						player.nickname, ' and juked' if juke_maps else '', mx_info['Name'], mx_info['Username'], self.site_short_name
					)
					await self.instance.chat(message)
				else:
					raise Exception('Unknown error while adding the map!')
			except Exception as e:
				logger.warning('Error when player {} was adding map from {}: {}'.format(player.login, self.site_short_name, str(e)))
				message = '$ff0Error: Can\'t add map {}, Error: {}'.format(mx_info['Name'], str(e))
				await self.instance.chat(message, player.login)

		# Save match settings after inserting maps.
		try:
			await self.instance.map_manager.save_matchsettings()
		except:
			pass

		# Reindex and create maps in database.
		try:
			await self.instance.map_manager.update_list(full_update=True)
		except:
			pass

		# Jukebox all the maps requested, in order.
		if juke_maps and len(added_map_uids) > 0:
			# Fetch map objects.
			for juke_uid in added_map_uids:
				map_instance = await self.instance.map_manager.get_map(uid=juke_uid)
				if map_instance:
					self.instance.apps.apps['jukebox'].insert_map(player, map_instance)

		return [await self.instance.map_manager.get_map(uid=added_uid) for added_uid in added_map_uids]



import asyncio
import uuid
import logging

from asyncio import iscoroutinefunction

from pyplanet.core.events import SignalManager
from pyplanet.core.ui.exceptions import ManialinkMemoryLeakException
from pyplanet.core.ui.template import Template

logger = logging.getLogger(__name__)


class _ManiaLink:
	def __init__(
		self, manager=None, id=None, version='3', body=None, template=None, timeout=0, hide_click=False, data=None,
		player_data=None, disable_alt_menu=False, throw_exceptions=False, relaxed_updating=False,
	):
		"""
		Create manialink (USE THE MANAGER CREATE, DONT INIT DIRECTLY!

		:param manager: Manager instance. use your app manager.
		:param id: Unique manialink id. Could be set later, must be set before displaying.
		:param version: Version of manialink.
		:param body: Body of manialink, not including manialink tags!!
		:param template: Template instance.
		:param timeout: Timeout to display, hide after the timeout is reached. Seconds.
		:param hide_click: Hide manialink when click is fired on button.
		:param data: Data to render. Could also be set later on or controlled separate from this instance.
		:param player_data: Dict with player login and for value the player specific variables. Dont fill this to have
		a global manialink instead of per person.
		:param throw_exceptions: Throw exceptions during handling and executing of action handlers.
		:param relaxed_updating: Relaxed updating will rate limit the amount of updates send to clients.
		:type manager: pyplanet.core.ui.AppUIManager
		:type template: pyplanet.core.ui.template.Template
		:type id: str
		:type version: str
		:type timeout: int
		"""
		self.manager = manager
		self.id = id or uuid.uuid4().hex
		self.version = version
		self.body = body
		self._template = template
		self.timeout = timeout
		self.hide_click = hide_click
		self.data = data if data and isinstance(data, dict) else dict()
		self.player_data = player_data if player_data and isinstance(player_data, dict) else dict()
		self.throw_exceptions = False
		self.disable_alt_menu = bool(disable_alt_menu)
		self.relaxed_updating = relaxed_updating

		self.receivers = dict()
		self._is_global_shown = False
		self._is_player_shown = dict()  # Holds per player login a boolean if the ml is shown.

		self.__register_listener = False

	async def is_global(self):
		return not self.player_data or self.player_data.keys() == 0

	async def get_template(self):
		return self._template

	async def render(self, player_login=None, data=None, player_data=None, template=None):
		"""
		Render template. Will render template and return body.

		:param player_login: Render data only for player, set to None to globally render (and ignore player_data).
		:param data: Data to append.
		:param player_data: Data to append.
		:param template: Template instance to use.
		:type template: pyplanet.core.ui.template.Template
		:return: Body, rendered manialink + script.
		"""
		if data and isinstance(data, dict):
			self.data.update(data)
		if not player_data:
			player_data = self.player_data or dict()
		if template and isinstance(template, Template):
			self._template = template
		if not template:
			template = await self.get_template()
		if not isinstance(template, Template):
			raise Exception('Can\'t render, no template is given!')

		# PyPlanet Global data.
		global_data = dict(
			_game=self.manager.instance.game,
			_instance=self.manager.instance,
			_app=self.manager.app if hasattr(self.manager, 'app') else None,
		)

		# Combine data (global + user specific).
		payload_data = global_data
		payload_data.update(self.data.copy())
		if player_login:
			payload_data.update(player_data.get(player_login, dict()))

		# Render and save in content.
		return await template.render(**payload_data)

	async def display(self, player_logins=None, **kwargs):
		"""
		Display the manialink. Will also render if no body is given. Will show per player or global. depending on
		the data given and stored!

		:param player_logins: Only display to the list of player logins given.
		"""
		if player_logins:
			for login in player_logins:
				self._is_player_shown[login] = True
		else:
			self._is_global_shown = True

		if not self.__register_listener:
			# Register handle
			SignalManager.listen('maniaplanet:manialink_answer', self.handle)
			self.__register_listener = True

		return await self.manager.send(self, player_logins, **kwargs)

	async def hide(self, player_logins=None):
		"""
		Hide manialink globally of only for the logins given in parameter.

		:param player_logins: Only hide for list of players, None for all players on the server.
		"""
		if player_logins:
			for login in player_logins:
				try:
					del self._is_player_shown[login]
				except:
					pass
		else:
			self._is_global_shown = False

		return await self.manager.hide(self, player_logins)

	def subscribe(self, action, target):
		"""
		Subscribe to a action given by the manialink.

		:param action: Action name.
		:param target: Target method.
		:return:
		"""
		if action not in self.receivers:
			self.receivers[action] = list()
		self.receivers[action].append(target)

	async def handle(self, player, action, values, **kwargs):
		if not action.startswith(self.id):
			return

		if not self._is_global_shown and player.login not in self._is_player_shown.keys():
			# Ignore if id is unique (uuid4)
			try:
				uuid.UUID(self.id, version=4)
			except:
				raise ManialinkMemoryLeakException(
					'Old view instance (ml-id: {}) is not yet destroyed, but is receiving player callbacks!, '
					'Make sure you are not removing old view instances with .destroy() and del variable! '
					'Potential Memory Leak!! Should be fixed asap!'.format(self.id)
				)

		action_name = action[len(self.id)+2:]
		if action_name not in self.receivers:
			return await self.handle_catch_all(player, action_name, values)

		# Call receivers.
		for rec in self.receivers[action_name]:
			try:
				if iscoroutinefunction(rec):
					await rec(player, action, values)
				else:
					rec(player, action, values)
			except Exception as e:
				if self.throw_exceptions:
					raise
				else:
					logging.exception('Exception has been silenced in ManiaLink Action receiver:', exc_info=e)

	async def handle_catch_all(self, player, action, values, **kwargs):
		"""
		Override this class to handle all other actions related to this view/manialink.

		:param player: Player instance.
		:param action: Action name/string
		:param values: Values provided by the user client.
		:param kwargs: *
		"""
		pass

	async def destroy(self):
		"""
		Destroy the Manialink with it's handlers and references.
		Will also hide the Manialink for all users!
		"""
		try:
			SignalManager.get_signal('maniaplanet:manialink_answer').unregister(self.handle)
		except Exception as e:
			logging.exception(e)
		try:
			await self.manager.destroy(self)
		except:
			pass
		self.receivers = dict()
		self.data = None
		self.player_data = None

	def destroy_sync(self):
		"""
		Destroy the Manialink with it's handlers and references.
		Will also hide the Manialink for all users!

		This method is sync and will call a async method (destroying of the manialink at our players) async but will not
		be executed at the same time. Be aware with this one!
		"""
		try:
			SignalManager.get_signal('maniaplanet:manialink_answer').unregister(self.handle)
			asyncio.ensure_future(self.manager.destroy(self))
		except Exception as e:
			logging.exception(e)
		self.receivers = dict()
		self.data = None
		self.player_data = None

	def __del__(self):
		self.destroy_sync()


class StaticManiaLink(_ManiaLink):
	"""
	The StaticManiaLink is mostly used in PyPlanet for general views. Please use the ``View`` classes instead of this
	core ui component!
	"""
	pass


class DynamicManiaLink(_ManiaLink):
	"""
	The DynamicManiaLink is a special manialink with data-bindings and automatically updates via maniascript.
	Please use the ``View`` classes instead!

	.. warning ::

		This feature is not yet implemented.

	"""
	def __init__(self, id):
		super().__init__(id)
		raise NotImplementedError



def toposort(graph):
	"""
	Perform a topological sort on a graph
	This returns an ordering of the nodes in the graph that places all
	dependencies before the nodes that require them.
	
	:param graph: an adjacency dict {node1: [dep1, dep2], node2: [dep1, dep3]}
	:return: A list of ordered nodes
	:rtype: list
	"""
	result = []
	used = set()

	def use(v, top):
		if v in used:
			return

		for parent in graph.get(v, []):
			if parent is top:
				raise ValueError("graph is cyclical through", parent)

			use(parent, v)

		used.add(v)
		result.append(v)

	for v in graph:
		use(v, v)

	return result



"""
The commands contributed package contains command management and callback logic.
"""
from .manager import CommandManager
from .command import Command
from .params import ParameterParser

__all__ = [
	'CommandManager',
	'Command',
	'ParameterParser',
]



"""
Map Admin methods and functions.
"""
import asyncio
import logging
from argparse import Namespace
from random import shuffle

from pyplanet.apps.core.maniaplanet.models import Map
from pyplanet.conf import settings
from pyplanet.contrib.command import Command
from pyplanet.contrib.map.exceptions import ModeIncompatible
from pyplanet.contrib.setting import Setting
from pyplanet.utils import gbxparser
from pyplanet.views.generics import ask_confirmation

logger = logging.getLogger(__name__)


class MapAdmin:
	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.admin.app.Admin
		"""
		self.app = app
		self.instance = app.instance

		self.setting_juke_after_adding = Setting(
			'juke_after_adding', 'Juke map after adding to the server', Setting.CAT_BEHAVIOUR, type=bool,
			description='Add the map just added from file or MX to the jukebox.',
			default=True
		)

	async def on_start(self):
		await self.instance.permission_manager.register('previous', 'Skip to the previous map', app=self.app, min_level=1)
		await self.instance.permission_manager.register('next', 'Skip to the next map', app=self.app, min_level=1)
		await self.instance.permission_manager.register('restart', 'Restart the maps', app=self.app, min_level=1)
		await self.instance.permission_manager.register('replay', 'Replay the maps', app=self.app, min_level=1)
		await self.instance.permission_manager.register('extend', 'Extend the TA limit', app=self.app, min_level=1)
		await self.instance.permission_manager.register('add_local_map', 'Add map from server disk', app=self.app, min_level=2)
		await self.instance.permission_manager.register('remove_map', 'Remove map from server', app=self.app, min_level=2)
		await self.instance.permission_manager.register('write_map_list', 'Write Matchsettings to file', app=self.app, min_level=2)
		await self.instance.permission_manager.register('read_map_list', 'Read and load specific Matchsettings file', app=self.app, min_level=2)
		await self.instance.permission_manager.register('shuffle', 'Shuffle map list order', app=self.app, min_level=2)

		await self.app.context.setting.register(self.setting_juke_after_adding)

		await self.instance.command_manager.register(
			Command(command='next', target=self.next_map, perms='admin:next', admin=True,
					description='Skip this map immediately.'),
			Command(command='skip', target=self.next_map, perms='admin:next', admin=True,
					description='Skip this map immediately.'),
			Command(command='previous', aliases=['prev'], target=self.prev_map, perms='admin:previous', admin=True,
					description='Switch the previously played map.'),
			Command(command='restart', aliases=['res', 'rs'], target=self.restart_map, perms='admin:restart', admin=True,
					description='Restart the current map immediately.'),
			Command(command='replay', target=self.replay_map, perms='admin:replay', admin=True,
					description='Replay the current map, adds it to the end of the jukebox.'),
			Command(command='local', namespace='add', target=self.add_local_map, perms='admin:add_local_map', admin=True,
					description='Add provided map file from local disk to the maplist.')
				.add_param('map', nargs=1, type=str, required=True, help='Map filename (relative to Maps directory).'),
			Command(command='remove', target=self.remove_map, perms='admin:remove_map', admin=True,
					description='Remove map from maplist.')
				.add_param('nr', required=False, type=int, help='The number from a list window or the unique identifier.'),
			Command(command='erase', target=self.erase_map, perms='admin:remove_map', admin=True,
					description='Remove and delete map from maplist and disk.')
				.add_param('nr', required=False, type=int, help='The number from a list window or the unique identifier.'),
			Command(command='writemaplist', aliases=['wml'], target=self.write_map_list, perms='admin:write_map_list', admin=True,
					description='Write the current maplist to the file on disk.')
				.add_param('file', required=False, type=str, help='Give custom match settings file to save to.'),
			Command(command='readmaplist', aliases=['rml'], target=self.read_map_list, perms='admin:read_map_list', admin=True,
					description='Read the maplist from the file on disk.')
				.add_param('file', required=True, type=str, help='Give custom match settings file to load from.'),
			Command(command='shuffle', target=self.shuffle, perms='admin:shuffle', admin=True,
					description='Shuffle the maplist.'),
			Command(command='extend', target=self.extend, perms='admin:extend', admin=True,
					description='Extend the playing time on the current map.')
				.add_param('seconds', required=False, type=int, help='Extend the TA limit with given seconds.'),
		)

		# If jukebox app is loaded, register the map actions.
		if 'jukebox' in self.instance.apps.apps:
			from pyplanet.apps.contrib.jukebox.views import MapListView
			MapListView.add_action(self.list_action_remove, 'Delete', '&#xf1f8;')

	async def list_action_remove(self, player, values, map_dictionary, view, **kwargs):
		# Check permission.
		if not await self.instance.permission_manager.has_permission(player, 'admin:remove_map'):
			await self.instance.chat(
				'$f00You don\'t have the permission to perform this action!',
				player
			)
			return

		# Ask for confirmation.
		cancel = bool(await ask_confirmation(player, 'Are you sure you want to remove the map \'{}\'$z$s from the server?'.format(
			map_dictionary['name']
		), size='sm'))
		if cancel is True:
			return

		# Simulate command.
		await self.remove_map(player, Namespace(nr=map_dictionary['id']))

		# Remove from the cache.
		for item in view.cache:
			if item['id'] == map_dictionary['id']:
				view.cache.remove(item)

		# Reload parent view.
		await view.refresh(player)

	async def prev_map(self, player, data, **kwargs):
		if not self.instance.map_manager.previous_map:
			message = '$ff0Error: Previous map is not known'
			return await self.instance.chat(message, player.login)
		if self.instance.map_manager.previous_map == self.instance.map_manager.current_map:
			message = '$ff0Error: Previous map is the same as the current map'
			return await self.instance.chat(message, player.login)

		if 'jukebox' in self.instance.apps.apps:
			self.instance.apps.apps['jukebox'].insert_map(player, self.instance.map_manager.previous_map)
		else:
			await self.instance.map_manager.set_next_map(self.instance.map_manager.previous_map)

		message = '$ff0Admin $fff{}$z$s$ff0 has skipped to the previous map.'.format(player.nickname)
		await self.instance.gbx.multicall(
			self.instance.gbx('NextMap'),
			self.instance.chat(message)
		)

	async def next_map(self, player, data, **kwargs):
		message = '$ff0Admin $fff{}$z$s$ff0 has skipped to the next map.'.format(player.nickname)
		await self.instance.gbx.multicall(
			self.instance.gbx('NextMap'),
			self.instance.chat(message)
		)

	async def restart_map(self, player, data, **kwargs):
		message = '$ff0Admin $fff{}$z$s$ff0 has restarted the map.'.format(player.nickname)

		# Dedimania save vreplay/ghost replays first.
		if 'dedimania' in self.instance.apps.apps:
			logger.info('Saving dedimania (v)replays first!..')
			if hasattr(self.instance.apps.apps['dedimania'], 'podium_start'):
				try:
					await self.instance.apps.apps['dedimania'].podium_start()
				except Exception as e:
					logger.exception(e)

		await self.instance.gbx.multicall(
			self.instance.gbx('RestartMap'),
			self.instance.chat(message)
		)

	async def replay_map(self, player, data, **kwargs):
		if 'jukebox' in self.instance.apps.apps:
			self.instance.apps.apps['jukebox'].insert_map(player, self.instance.map_manager.current_map)
		else:
			await self.instance.map_manager.set_next_map(self.instance.map_manager.current_map)

		await self.instance.chat(
			'$ff0Admin $fff{}$z$s$ff0 has queued this map for replay.'.format(player.nickname)
		)

	async def write_map_list(self, player, data, **kwargs):
		setting = settings.MAP_MATCHSETTINGS
		if isinstance(setting, dict) and self.instance.process_name in setting:
			setting = setting[self.instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not setting and not data.file:
			message = '$ff0Default match settings file not configured in your settings!'
			return await self.instance.chat(message, player)
		if data.file:
			file_name = data.file
		else:
			file_name = setting.format(server_login=self.instance.game.server_player_login)

		file_path = 'MatchSettings/{}'.format(file_name)
		message = '$ff0Match Settings has been saved to the file: {}'.format(file_name)
		await self.instance.map_manager.save_matchsettings(file_path)

		# Send message + reload all maps in memory.
		await asyncio.gather(
			self.instance.chat(message, player),
			self.instance.map_manager.update_list(full_update=True)
		)

	async def read_map_list(self, player, data, **kwargs):
		file_name = data.file
		file_path = 'MatchSettings/{}'.format(file_name)

		try:
			await self.instance.map_manager.load_matchsettings(file_path)
			message = '$ff0Match Settings has been loaded from: {}'.format(file_path)
		except:
			message = '$ff0Could not load match settings! Does the file exists? Check log for details.'

		# Send message + reload all maps in memory.
		await asyncio.gather(
			self.instance.chat(message, player),
			self.instance.map_manager.update_list(full_update=True)
		)

	async def shuffle(self, player, data, **kwargs):
		# First, retrieve the current maplist.
		map_list = await self.instance.gbx('GetMapList', -1, 0)
		file_names = [m['FileName'] for m in map_list]

		# Remove all maps from the server.
		await self.instance.gbx('RemoveMapList', file_names)

		# Shuffle map file names.
		shuffle(file_names)

		# Re-add all maps in new order.
		await self.instance.gbx('AddMapList', file_names)

		# Send message + reload all maps in memory.
		await asyncio.gather(
			self.instance.chat('$ff0Maps have been shuffled!', player),
			self.instance.map_manager.update_list(full_update=True)
		)

	async def add_local_map(self, player, data, **kwargs):
		map_file = data.map

		# Check for the file.
		if not await self.instance.storage.driver.exists('UserData/Maps/{}'.format(
			map_file
		)):
			message = '$ff0Error: Can\'t add map because the file is not found!'
			await self.instance.chat(message, player.login)
			return

		# Fetch setting if juke after adding is enabled.
		juke_maps = await self.setting_juke_after_adding.get_value()
		if 'jukebox' not in self.instance.apps.apps:
			juke_maps = False
		juke_list = list()

		try:
			# Parse GBX file.
			async with self.instance.storage.open_map(map_file) as map_fh:
				parser = gbxparser.GbxParser(buffer=map_fh)
				map_info = await parser.parse()

			# Test if map isn't yet in our current map list.
			if self.instance.map_manager.playlist_has_map(map_info['uid']):
				raise Exception('Map already in playlist! Update? remove it first!')

			# Insert map to server.
			result = await self.instance.map_manager.add_map(map_file)

			if result:
				# Juke if setting has been provided.
				if juke_maps:
					juke_list.append(map_info['uid'])

				message = '$ff0Admin $fff{}$z$s$ff0 has added{} the map $fff{}$z$s$ff0 by $fff{}$z$s$ff0.'.format(
					player.nickname, ' and juked' if juke_maps else '', map_info['name'], map_info['author_nickname']
				)
				await self.instance.chat(message)
			else:
				raise Exception('Unknown error while adding the map!')

			# Save match settings after inserting maps.
			try:
				await self.instance.map_manager.save_matchsettings()
			except:
				pass

			# Reindex and create maps in database.
			try:
				await self.instance.map_manager.update_list(full_update=True)
			except:
				pass

			# Jukebox all the maps requested, in order.
			if juke_maps and len(juke_list) > 0:
				# Fetch map objects.
				for juke_uid in juke_list:
					map_instance = await self.instance.map_manager.get_map(uid=juke_uid)
					if map_instance:
						self.instance.apps.apps['jukebox'].insert_map(player, map_instance)

		except Exception as e:
			logger.warning('Error when player {} was adding map from local disk: {}'.format(player.login, str(e)))
			message = '$ff0Error: Can\'t add map, Error: {}'.format(str(e))
			await self.instance.chat(message, player.login)

	async def erase_map(self, player, data, **kwargs):
		kwargs['erase'] = True
		return await self.remove_map(player, data, **kwargs)

	async def remove_map(self, player, data, **kwargs):
		map_nr = getattr(data, 'nr', None)
		erase = kwargs.get('erase', False)

		try:
			# Make sure we get the map instance.
			if not map_nr:
				map_instance = self.instance.map_manager.current_map
			else:
				map_instance = await Map.get(id=map_nr)

			# Send remove command.
			await self.instance.map_manager.remove_map(map_instance, delete_file=erase)

			# Send message to all.
			message = '$ff0Admin $fff{}$z$s$ff0 has {} the map $fff{}$z$s$ff0.'.format(
				player.nickname, 'erased' if erase else 'removed', map_instance.name
			)
			await self.instance.chat(message)
		except Exception as e:
			# Handle errors.
			logger.error(str(e))
			message = '$ff0Error: Can\'t remove map, Error: {}'.format(str(e))
			await self.instance.chat(message, player)

	async def extend(self, player, data, **kwargs):
		extend_with = data.seconds
		try:
			extended_with = await self.instance.map_manager.extend_ta(extend_with=extend_with)
		except ModeIncompatible:
			return await self.instance.chat('$ff0Error: Game mode must be Time Attack to use the extend functionality!', player)

		message = '$ff0Admin $fff{}$z$s$ff0 has extended the time limit with $fff{} seconds.'.format(
			player.nickname, extended_with
		)
		await self.instance.chat(message)



"""
GBXRemote 2 client for python 3.5+ part of PyPlanet.
"""
import asyncio
import json
import logging
import struct

from xmlrpc.client import dumps, loads, Fault
from xml.parsers.expat import ExpatError

from pyplanet.core.exceptions import TransportException
from pyplanet.core.events.manager import SignalManager
from pyplanet.utils.log import handle_exception

logger = logging.getLogger(__name__)


class GbxRemote:
	"""
	The GbxClient holds the connection to the dedicated server. Maintains the queries and the handlers it got.
	"""
	MAX_REQUEST_SIZE  = 2000000  # 2MB
	MAX_RESPONSE_SIZE = 4000000  # 4MB

	def __init__(self, host, port, event_pool=None, user=None, password=None, api_version='2013-04-16', instance=None):
		"""
		Initiate the GbxRemote client.

		:param host: Host of the dedicated server.
		:param port: Port of the dedicated XML-RPC server.
		:param event_pool: Asyncio pool to execute the handling on.
		:param user: User to authenticate with, in most cases this is 'SuperAdmin'
		:param password: Password to authenticate with.
		:param api_version: API Version to use. In most cases you won't override the default because version changes
							should be abstracted by the other core components.
		:param instance: Instance of the app.
		:type host: str
		:type port: str int
		:type event_pool: asyncio.BaseEventPool
		:type user: str
		:type password: str
		:type api_version: str
		:type instance: pyplanet.core.instance.Instance
		"""
		self.host = host
		self.port = port
		self.user = user
		self.password = password
		self.api_version = api_version
		self.instance = instance

		self.dedicated_version = None
		self.dedicated_build = None

		self.event_loop = event_pool or asyncio.get_event_loop()
		self.gbx_methods = list()

		self.handlers = dict()
		self.handler_nr = 0x80000000

		self.script_handlers = dict()

		self.reader = None
		self.writer = None
		self.loop_task = None

	@classmethod
	def create_from_settings(cls, instance, conf):
		"""
		Create an instance from configuration given for the specific pool.

		:param instance: Instance of the app.
		:param conf: Settings for pool.
		:type conf: dict
		:return: Instance of XML-RPC GbxClient.
		:rtype: pyplanet.core.gbx.client.GbxClient
		"""
		return cls(
			instance=instance,
			host=conf['HOST'], port=conf['PORT'], user=conf['USER'], password=conf['PASSWORD']
		)

	def get_next_handler(self):
		handler = self.handler_nr
		if self.handler_nr == 0xffffffff:
			logger.debug('GBX: Reached max handler numbers, RESETTING TO ZERO!')
			self.handler_nr = 0x80000000
		else:
			self.handler_nr += 1
		return handler

	async def connect(self):
		"""
		Make connection to the server. This will first check the protocol version and after successful connection
		also authenticate, set the API version and enable callbacks.
		"""
		logger.debug('Trying to connect to the dedicated server...')

		# Create socket (+ retry few times if not successful.
		retries = 0
		max_retries = 10
		while True:
			try:
				self.reader, self.writer = await asyncio.open_connection(
					host=self.host,
					port=self.port,
					loop=self.event_loop,
				)
				break
			except Exception as exc:
				if retries >= max_retries:
					raise
				retries += 1

				logger.info('Couldn\'t connect to Dedicated Server. Retry {} of {} (Error: {}'.format(
					retries,
					max_retries,
					str(exc)
				))
				await asyncio.sleep(2)

		_, header = struct.unpack_from('<L11s', await self.reader.readexactly(15))
		if header.decode() != 'GBXRemote 2':
			raise TransportException('Server is not a valid GBXRemote 2 server.')
		logger.debug('Dedicated connection established!')

		# From now we need to start listening.
		self.loop_task = self.event_loop.create_task(self.listen())

		# Startup tasks.
		await self.execute('Authenticate', self.user, self.password)
		await asyncio.gather(
			self.execute('SetApiVersion', self.api_version),
			self.execute('EnableCallbacks', True),
		)

		# Fetch gbx methods and dedicated server version.
		self.gbx_methods, versions = await asyncio.gather(
			self.execute('system.listMethods'),
			self.execute('GetVersion')
		)
		self.dedicated_version = versions['Version']
		self.dedicated_build = versions['Build']
		#Make a check for SystemInformation if Dedicated = True
		system_info = await self.execute('GetSystemInfo')
		if system_info['IsDedicated'] == False and system_info['IsServer'] == False:
			logger.debug('Dedicated seems to be a gameclient! Adjust `PORT` in /settings/base.py')
			exit(50)
		
		# Check for scripted mode.
		mode = await self.execute('GetGameMode')
		settings = await self.execute('GetModeScriptSettings')
		if mode == 0:
			if 'S_UseScriptCallbacks' in settings:
				settings['S_UseScriptCallbacks'] = True
			if 'S_UseLegacyCallback' in settings:
				settings['S_UseLegacyCallback'] = False
			if 'S_UseLegacyXmlRpcCallbacks' in settings:
				settings['S_UseLegacyXmlRpcCallbacks'] = False
			await asyncio.gather(
				self.execute('SetModeScriptSettings', settings),
				self.execute('TriggerModeScriptEventArray', 'XmlRpc.EnableCallbacks', ['true'])
			)

		logger.debug('Dedicated authenticated, API version set and callbacks enabled!')

	async def disconnect(self):
		"""
		Stop the task of listening, destroy connections, reader and writer.
		"""
		if self.loop_task:
			self.loop_task.cancel()
			del self.loop_task
		if self.reader:
			del self.reader
		if self.writer:
			self.writer.close()
			del self.writer

	async def execute(self, method, *args, timeout=45.0):
		"""
		Query the dedicated server and return the results. This method is a coroutine and should be awaited on.
		The result you get will be a tuple with data inside (the response payload).

		:param method: Server method.
		:param args: Arguments.
		:param timeout: Wait for x seconds until future is returned. Default is 45 seconds.
		:type method: str
		:type args: any
		:return: Tuple with response data (after awaiting).
		:rtype: Future<tuple>
		"""
		request_bytes = dumps(args, methodname=method, allow_none=True).encode()
		length_bytes = len(request_bytes).to_bytes(4, byteorder='little')
		handler = self.get_next_handler()

		handler_bytes = handler.to_bytes(4, byteorder='little')

		# Create new future to be returned.
		self.handlers[handler] = future = asyncio.Future()

		# Send to server.
		self.writer.write(length_bytes + handler_bytes + request_bytes)

		return await asyncio.wait_for(future, timeout)

	async def listen(self):
		"""
		Listen to socket.
		"""
		try:
			while True:
				head = await self.reader.readexactly(8)
				size, handle = struct.unpack_from('<LL', head)
				body = await self.reader.readexactly(size)
				data = method = fault = None

				try:
					data, method = loads(body, use_builtin_types=True)
				except Fault as e:
					fault = e
				except ExpatError as e:
					# See #121 for this solution.
					handle_exception(exception=e, module_name=__name__, func_name='listen', extra_data={'body': body})
					continue

				if data and len(data) == 1:
					data = data[0]

				self.event_loop.create_task(self.handle_payload(handle, method, data, fault))
		except ConnectionResetError as e:
			logger.critical(
				'Connection with the dedicated server has been closed, we will now close down the subprocess! {}'.format(str(e))
			)
			# When the connection has been reset, we will close the controller process so it can be restarted by the god
			# process. Exit code 10 gives the information to the god process.
			exit(10)
		except Exception as e:
			handle_exception(exception=e, module_name=__name__, func_name='listen')
			raise

	async def handle_payload(self, handle_nr, method=None, data=None, fault=None):
		"""
		Handle a callback/response payload or fault.

		:param handle_nr: Handler ID
		:param method: Method name
		:param data: Parsed payload data.
		:param fault: Fault object.
		"""
		if handle_nr in self.handlers:
			await self.handle_response(handle_nr, method, data, fault)
		elif method and data is not None:
			if method == 'ManiaPlanet.ModeScriptCallbackArray':
				await self.handle_scripted(handle_nr, method, data)
			elif method == 'ManiaPlanet.ModeScriptCallback':
				await self.handle_scripted(handle_nr, method, data)
			else:
				await self.handle_callback(handle_nr, method, data)
		elif fault is not None:
			raise TransportException('Handle payload got invalid parameters, see fault exception! {}'.format(fault)) from fault
		else:
			print(method, handle_nr, data)
			logging.warning('Received gbx data, but handle wasn\'t known or payload invalid: handle_nr: {}, method: {}'.format(
				handle_nr, method,
			))

	async def handle_response(self, handle_nr, method=None, data=None, fault=None):
		logger.debug('GBX: Received response to handler {}, method: {}'.format(handle_nr, method))
		handler = self.handlers.pop(handle_nr)
		if not fault:
			handler.set_result(data)
		else:
			handler.set_exception(fault)
		handler.done()

	async def handle_callback(self, handle_nr, method, data):
		logger.debug('GBX: Received callback: {}: {}'.format(method, data))
		signal = SignalManager.get_callback(method)
		if signal:
			await signal.send_robust(data)

	async def handle_scripted(self, handle_nr, method, data):
		# Unpack first.
		method, raw = data

		# Check if we only have one response array length, mostly this is the case due to the gbx handling.
		# Only when we don't get any response we don't have this!
		if len(raw) == 1:
			raw = raw[0]

		# Show warning when using non-supported modes.
		try:
			if 'LibXmlRpc' in method:
				logger.warning('You are using an older gamemode script that isn\'t supported by PyPlanet (usage of LibXmlRpc_)')
		except:
			pass

		# Try to parse JSON, mostly the case.
		try:
			if isinstance(raw, list):
				payload = dict()
				for idx, part in enumerate(raw):
					try:
						payload.update(json.loads(part))
					except:
						payload['raw_{}'.format(idx)] = part
			else:
				payload = json.loads(raw)
		except Exception as e:
			payload = raw

		# Check if payload contains a responseid, when it does, we call the scripted handler future object.
		if isinstance(payload, dict) and 'responseid' in payload and len(payload['responseid']) > 0:
			response_id = payload['responseid']

			if response_id in self.script_handlers:
				logger.debug('GBX: Received scripted response to method: {} and responseid: {}'.format(method, response_id))
				handler = self.script_handlers.pop(response_id)
				handler.set_result(payload)
				handler.done()
				return
			else:
				# We don't have this handler registered, throw warning in console.
				logger.warning('GBX: Received scripted response with responseid, but no hander was registered! Payload: {}'.format(payload))
				return

		# If not, we should just throw it as an ordinary callback.
		logger.debug('GBX: Received scripted callback: {}: {}'.format(method, payload))

		signal = SignalManager.get_callback('Script.{}'.format(method))
		if signal:
			await signal.send_robust(payload)



from peewee import *

from pyplanet.core.db import Model


class Setting(Model):
	"""
	DON'T USE THIS DIRECTLY, USE THE CONTRIB MANAGER!
	"""

	app = CharField(
		max_length=150,
		null=True,
		index=True,
		default=None,
	)
	"""
	App label of setting. None for core (never use None for this outside of the manager itself!!).
	"""

	key = CharField(
		max_length=150,
		null=False,
		index=True
	)
	"""
	The key of the setting, together with app unique!
	"""

	category = CharField(
		max_length=100,
		null=True,
		default=None
	)
	"""
	Name of category, all items will be under a special label or menu item.
	"""

	name = CharField(
		max_length=255,
		null=False,
		help_text='Name of setting.'
	)
	"""
	Name of the setting. Will be displayed in a list.
	"""

	description = TextField(
		null=True,
		default=None,
		help_text='Description of the setting.'
	)
	"""
	Description of the setting contains the information to help the user with filling in the field(s).
	"""

	value = TextField(
		null=True,
		default=None,
	)
	"""
	The value, fitted in a text field so we can parse it quite easy.
	"""

	class Meta:
		indexes = (
			(('app', 'key'), True),
		)

	def __str__(self):
		return '{} ({}:{})'.format(self.app, self.key, self.name)

	@classmethod
	async def get_or_create_from_info(cls, key, app, **kwargs):
		"""
		This method will be called from the core, getting or creating a setting instance from the information we got from
		the setting instance (static setting object).
		
		:param key: Key of setting.
		:param app: App label.
		:param kwargs: Other key arguments, matching the model columns!
		:return: Setting instance.
		"""
		needs_save = False
		try:
			setting = await cls.get(key=key, app=app)
		except DoesNotExist:
			setting = cls(key=key, app=app)
			needs_save = True

		# Update from the kwargs.
		for k, v in kwargs.items():
			if v is not None and hasattr(setting, k) and getattr(setting, k) != v:
				setattr(setting, k, v)
				needs_save = True

		if needs_save:
			await setting.save()

		return setting



"""
This file contains the core signals, related to the controller itself.
"""
from pyplanet.core.events import Signal as _Signal
from pyplanet.core.events.manager import SignalManager as _SignalManager

pyplanet_start_before		= _Signal(code='start_before', namespace='pyplanet')
pyplanet_start_after		= _Signal(code='start_after', namespace='pyplanet')

pyplanet_start_gbx_before	= _Signal(code='start_gbx_before', namespace='pyplanet')
pyplanet_start_gbx_after	= _Signal(code='start_gbx_after', namespace='pyplanet')

pyplanet_start_db_before	= _Signal(code='start_db_before', namespace='pyplanet')
pyplanet_start_db_after		= _Signal(code='start_db_after', namespace='pyplanet')

pyplanet_start_apps_before	= _Signal(code='start_apps_before', namespace='pyplanet')
pyplanet_start_apps_after	= _Signal(code='start_apps_after', namespace='pyplanet')

pyplanet_performance_mode_begin = _Signal(code='start_performance_mode', namespace='pyplanet')
pyplanet_performance_mode_end	= _Signal(code='end_performance_mode', namespace='pyplanet')

_SignalManager.register_signal([
	pyplanet_start_before, pyplanet_start_after, pyplanet_start_gbx_before ,pyplanet_start_gbx_after,
	pyplanet_start_db_before, pyplanet_start_db_after, pyplanet_start_apps_before, pyplanet_start_apps_after,
	pyplanet_performance_mode_begin, pyplanet_performance_mode_end
])



"""
This file contains the static entry point to access the controller instance from anywhere without having a reference
to it and to prevent issues with circular imports.

Please import this from the ``pyplanet.core`` package instead!
"""


class _Controller:
	def __init__(self, *args, **kwargs):
		self.name = None
		self.__instance = None

	def prepare(self, name):
		from pyplanet.core.instance import Instance

		self.name = name
		self.__instance = Instance(name)
		return self

	@property
	def instance(self):
		"""
		Get active instance in current process.

		:return: Controller Instance
		:rtype: pyplanet.core.instance.Instance
		"""
		return self.__instance

Controller = _Controller()
"""
Controller access point to prevent circular imports. This is a lazy provided way to get the instance from anywhere!
"""



from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.apps.core.statistics.models import Score
from pyplanet.utils import times
from pyplanet.views.generics.list import ManualListView


class StatsScoresListView(ManualListView):
	title = 'Personal time progression on this map'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app, player):
		"""
		Init score list view.

		:param player: Player instance.
		:param app: App instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		super().__init__(self)

		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.provide_search = False

	async def get_data(self):
		score_list = await Score.objects.execute(
			Score.select(Score, Player)
				.join(Player)
				.where(Score.map == self.app.instance.map_manager.current_map.get_id())
				.order_by(Score.created_at.asc())
		)

		scores_list = list(score_list)

		personal_list = [s for s in scores_list if s.player.id == self.player.get_id()]

		if len(personal_list) == 0:
			message = '$i$f00There are no personal scores available for $fff{}$z$s$f00$i!'.format(self.app.instance.map_manager.current_map.name)
			await self.app.instance.chat(message, self.player)
			return

		local_record = min([s.score for s in scores_list])

		scores = list()
		last_best = 0
		last_best_index = 1
		personal_best = min([s.score for s in personal_list])
		for score_in_list in personal_list:
			historical_local = min([s.score for s in scores_list if s.created_at <= score_in_list.created_at])

			score = dict()
			score['index'] = ''
			score['score'] = times.format_time(score_in_list.score)
			score['created_at'] = score_in_list.created_at.strftime('%d-%m-%Y %H:%M:%S')
			score['difference_to_pb'] = times.format_time((score_in_list.score - personal_best))
			score['difference_to_prev'] = ''
			score['difference_to_local'] = times.format_time((score_in_list.score - local_record))
			score['historical_local'] = times.format_time(historical_local)
			score['difference_to_hist_local'] = times.format_time((score_in_list.score - historical_local))

			if last_best == 0:
				score['index'] = last_best_index
				last_best = score_in_list.score
				last_best_index += 1
			elif score_in_list.score < last_best:
				score['index'] = last_best_index
				score['difference_to_prev'] = '$00f- {}'.format(times.format_time((last_best - score_in_list.score)))
				last_best = score_in_list.score
				last_best_index += 1
			else:
				score['difference_to_prev'] = '$f00+ {}'.format(times.format_time((score_in_list.score - last_best)))

			if score_in_list.score == local_record:
				score['difference_to_local'] = ''
			else:
				score['difference_to_local'] = '$f00+ {}'.format(score['difference_to_local'])

			if score_in_list.score == historical_local:
				score['difference_to_hist_local'] = ''
			else:
				score['difference_to_hist_local'] = '$f00+ {}'.format(score['difference_to_hist_local'])

			if score_in_list.score == personal_best:
				score['index'] = 'PB'
				score['difference_to_pb'] = ''
			else:
				score['difference_to_pb'] = '$f00+ {}'.format(score['difference_to_pb'])

			scores.append(score)

		return scores

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'index',
				'sorting': False,
				'searching': False,
				'width': 6,
				'type': 'label'
			},
			{
				'name': 'Time',
				'index': 'score',
				'sorting': False,
				'searching': False,
				'width': 25,
				'type': 'label'
			},
			{
				'name': 'From PB',
				'index': 'difference_to_pb',
				'sorting': False,
				'searching': False,
				'width': 28,
				'type': 'label'
			},
			{
				'name': 'From prev. PB',
				'index': 'difference_to_prev',
				'sorting': False,
				'searching': False,
				'width': 34,
				'type': 'label'
			},
			{
				'name': 'From Local',
				'index': 'difference_to_local',
				'sorting': False,
				'searching': False,
				'width': 30,
				'type': 'label'
			},
			{
				'name': 'Hist. Local',
				'index': 'historical_local',
				'sorting': False,
				'searching': False,
				'width': 28,
				'type': 'label'
			},
			{
				'name': 'From hist. Local',
				'index': 'difference_to_hist_local',
				'sorting': False,
				'searching': False,
				'width': 34,
				'type': 'label'
			},
			{
				'name': 'Driven at',
				'index': 'created_at',
				'sorting': False,
				'searching': False,
				'width': 40,
				'type': 'label'
			},
		]


class CheckpointComparisonView(ManualListView):
	title = 'Checkpoint comparison on this map'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app, player):
		"""
		Init score list view.

		:param player: Player instance.
		:param app: App instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		super().__init__(self)

		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.provide_search = False

	async def get_data(self):
		score_list = await Score.objects.execute(
			Score.select(Score, Player)
				.join(Player)
				.where(Score.map == self.app.instance.map_manager.current_map.get_id())
				.order_by(Score.score.asc())
		)

		scores_list = list(score_list)
		for score in scores_list:
			score_checkpoints = score.checkpoints.split(',')
			score.checkpoints = [int(y) - int(x) for x, y in zip(score_checkpoints, score_checkpoints[1:])]  # TODO: Improve performance of this code block.
			score.checkpoints.insert(0, int(score_checkpoints[0]))

		personal_list = [s for s in scores_list if s.player.id == self.player.get_id()]

		local_record = next(iter(scores_list or []), None)
		personal_best = next(iter(personal_list or []), None)

		checkpoints = list()
		total_pb = 0
		total_local = 0
		total_ideal = 0

		for checkpoint_index in range(0, self.app.instance.map_manager.current_map.num_checkpoints):
			pb_checkpoint = personal_best.checkpoints[checkpoint_index] if personal_best is not None else None
			local_checkpoint = local_record.checkpoints[checkpoint_index] if local_record is not None else None
			ideal_checkpoint = min([cp.checkpoints[checkpoint_index] for cp in scores_list]) if scores_list is not None and len(scores_list) > 0 else None
			ideal_record = [score for score in scores_list if score.checkpoints[checkpoint_index] == ideal_checkpoint][0] if scores_list is not None and len(scores_list) > 0 else None

			if pb_checkpoint is not None:
				total_pb += pb_checkpoint

			if local_checkpoint is not None:
				total_local += local_checkpoint
				total_ideal += ideal_checkpoint

			checkpoint = dict()
			checkpoint['index'] = (checkpoint_index + 1) if checkpoint_index < (self.app.instance.map_manager.current_map.num_checkpoints - 1) else 'Finish'
			checkpoint['personal_best'] = '{} ({})'.format(times.format_time(total_pb), times.format_time(pb_checkpoint)) if pb_checkpoint is not None else '-'
			checkpoint['local_record'] = '{} ({})'.format(times.format_time(total_local), times.format_time(local_checkpoint)) if local_checkpoint is not None else '-'
			checkpoint['local_record_driver'] = local_record.player.nickname if local_checkpoint is not None else '-'
			checkpoint['difference_to_local'] = '-'
			checkpoint['ideal'] = '{} ({})'.format(times.format_time(total_ideal), times.format_time(ideal_checkpoint)) if local_checkpoint is not None else '-'
			checkpoint['ideal_driver'] = ideal_record.player.nickname if local_checkpoint is not None else '-'
			checkpoint['difference_to_ideal'] = '-'

			if pb_checkpoint is not None:
				if total_pb > total_local:
					checkpoint['difference_to_local'] = '$f00+ {}'.format(times.format_time((total_pb - total_local)))
				else:
					checkpoint['difference_to_local'] = '$00f- {}'.format(times.format_time((total_local - total_pb)))

				if total_pb > total_ideal:
					checkpoint['difference_to_ideal'] = '$f00+ {}'.format(times.format_time((total_pb - total_ideal)))
				else:
					checkpoint['difference_to_ideal'] = '$00f- {}'.format(times.format_time((total_ideal - total_pb)))

			checkpoints.append(checkpoint)

		return checkpoints

	async def get_fields(self):
		return [
			{
				'name': 'CP #',
				'index': 'index',
				'sorting': False,
				'searching': False,
				'width': 15,
				'type': 'label'
			},
			{
				'name': 'Your PB',
				'index': 'personal_best',
				'sorting': False,
				'searching': False,
				'width': 32,
				'type': 'label'
			},
			{
				'name': 'Local Record',
				'index': 'local_record',
				'sorting': False,
				'searching': False,
				'width': 30,
				'type': 'label'
			},
			{
				'name': 'Driver',
				'index': 'local_record_driver',
				'sorting': False,
				'searching': False,
				'width': 30,
				'type': 'label'
			},
			{
				'name': 'From Local',
				'index': 'difference_to_local',
				'sorting': False,
				'searching': False,
				'width': 26,
				'type': 'label'
			},
			{
				'name': 'Ideal',
				'index': 'ideal',
				'sorting': False,
				'searching': False,
				'width': 30,
				'type': 'label'
			},
			{
				'name': 'Driver',
				'index': 'ideal_driver',
				'sorting': False,
				'searching': False,
				'width': 30,
				'type': 'label'
			},
			{
				'name': 'From Ideal',
				'index': 'difference_to_ideal',
				'sorting': False,
				'searching': False,
				'width': 26,
				'type': 'label'
			}
		]



"""
Queue View.
"""
from pyplanet.apps.contrib.queue.exception import PlayerAlreadyInQueue
from pyplanet.views import TemplateView


class QueueView(TemplateView):
	"""
	Queue View
	"""
	template_name = 'queue/queue.xml'

	def __init__(self, app):
		"""
		Initiate the queue view.

		:param app: Queue app instance.
		:type app: pyplanet.apps.contrib.queue.Queue
		"""
		super().__init__(app.context.ui)

		self.app = app
		self.id = 'queue_widget'

		self.subscribe('enter', self.enter_queue)
		self.subscribe('exit', self.exit_queue)

	async def enter_queue(self, player, *_, **__):
		try:
			await self.app.enter_queue(player)
			await self.app.slot_change()
		except PlayerAlreadyInQueue:
			pass

	async def exit_queue(self, player, *_, **__):
		await self.app.exit_queue(player)

	async def get_context_data(self):
		context = await super().get_context_data()
		context['total'] = await self.app.list.count()
		return context

	async def get_per_player_data(self, login):
		try:
			player = await self.app.instance.player_manager.get_player(login=login)
			return dict(
				position=await self.app.list.get_position(player)
			)
		except:
			return dict(
				position=None
			)



import logging
import logging.config
import os
import sys
import traceback

from raven import Client
from logging.handlers import QueueHandler as BaseQueueHandler

from pyplanet import __version__ as version
from pyplanet.conf import settings
from pyplanet.core.exceptions import ImproperlyConfigured


IGNORED_PATHS = [
	'aiomysql',
]
IGNORED_TEXT = [
	'InternalError: Packet sequence number wrong',
	'Lost connection to',
	'MemoryError',
]


class Raven:  # pragma: no cover
	CLIENT = None

	@classmethod
	def get_client(cls):
		if not cls.CLIENT:
			cls.CLIENT = Client(
				dsn='https://bcbafd2d81234d53b32c66077ae0a008:98dcc46acc484eeb95ebf9f0c30e6dd4@sentry.s3.toffe.me/2',
				environment='development' if settings.DEBUG else 'production',
				release=version,
				include_paths=['pyplanet']
			)
		return cls.CLIENT


def initiate_logger():  # pragma: no cover
	if settings.LOGGING_CONFIG == 'logging.config.dictConfig':
		logging.config.dictConfig(settings.LOGGING)

	# If we have writing logs enabled.
	if settings.LOGGING_WRITE_LOGS:
		# Parse directory
		if settings.LOGGING_DIRECTORY:
			if os.path.isabs(settings.LOGGING_DIRECTORY):
				path = settings.LOGGING_DIRECTORY
			else:
				path = os.path.join(settings.ROOT_PATH, settings.LOGGING_DIRECTORY)

			# Create if not exists!
			if not os.path.exists(path):
				os.mkdir(path)

			path = os.path.join(path, 'pyplanet.log')
		else:
			path = os.path.join(settings.ROOT_PATH, 'pyplanet.log')

		# Determinate handler and initiate it.
		if settings.LOGGING_ROTATE_LOGS:
			handler = logging.handlers.TimedRotatingFileHandler(
				path, when='D', interval=1, backupCount=14, encoding='utf-8'
			)
		else:
			handler = logging.FileHandler(
				path, encoding='utf-8'
			)

		# Change the formatter.
		handler.setFormatter(
			logging.Formatter('[%(asctime)s][%(levelname)s][%(threadName)s] %(name)s: %(message)s (%(filename)s:%(lineno)d)')
		)

		# Add to the root handler.
		logging.root.addHandler(handler)


def handle_exception(exception=None, module_name=None, func_name=None, extra_data=None, force=False):  # pragma: no cover
	# Test for ignored exceptions
	if exception and isinstance(exception, (ImproperlyConfigured,)):
		return

	from pyplanet.core import Controller
	if settings.DEBUG or settings.LOGGING_REPORTING == 0:
		if exception:
			logging.exception(exception)
		return

	# Filter out exceptions.
	ignore = False
	try:
		stack = traceback.extract_stack()
		for frame in stack:
			if any(ig in frame.filename for ig in IGNORED_PATHS):
				ignore = True
				break
	except:
		pass
	try:
		if not force:
			if any(ig.lower() in str(exception).lower() for ig in IGNORED_TEXT):
				ignore = True
	except:
		pass

	if ignore:
		return

	# Extra Data.
	if not extra_data:
		extra_data = dict()
	extra_data = extra_data.copy()
	if Controller.instance and Controller.instance.game:
		extra_data.update(dict(game=Controller.instance.game.__dict__))

	if exception and hasattr(exception, '__dict__'):
		extra_data['report_exception'] = exception.__dict__
	if module_name:
		extra_data['report_module_name'] = module_name
	if func_name:
		extra_data['report_func_name'] = func_name

	if settings.LOGGING_REPORTING == 3:
		extra_data['report_privacy'] = 'allow-share-contrib-apps'
	else:
		extra_data['report_privacy'] = 'deny-share-contrib-apps'

	if settings.LOGGING_REPORTING >= 2:
		Raven.get_client().extra_context(extra_data)

	# Send to sentry.
	exc_info = sys.exc_info()
	Raven.get_client().captureException(exc_info=exc_info)


class RequireDebugFalse(logging.Filter):
	def filter(self, record):
		return not settings.DEBUG


class RequireDebugTrue(logging.Filter):
	def filter(self, record):
		return settings.DEBUG


class RequireException(logging.Filter):
	def filter(self, record):
		return bool(record.exc_info)


class QueueHandler(BaseQueueHandler):  # pragma: no cover
	def prepare(self, record):
		# Override due to bug
		self.format(record)
		record.msg = record.msg or record.message
		record.args = None
		record.exc_info = None
		return record



import asynctest

from pyplanet.core import Controller
from pyplanet.views.generics import AlertView
from pyplanet.views.generics.alert import PromptView


class TestGenericViews(asynctest.TestCase):

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.instance = Controller.prepare(name='default').instance

	async def setUp(self):
		await self.instance.apps.discover()

	async def test_alert(self):
		view = AlertView(message='TestMessage', size='sm')
		body = await view.render()
		assert 'TestMessage' in body

		view = AlertView(message='TestMessage', size='md')
		body = await view.render()
		assert 'TestMessage' in body

		view = AlertView(message='TestMessage', size='lg')
		body = await view.render()
		assert 'TestMessage' in body

	async def test_prompt(self):
		view = PromptView(message='TestMessage', size='sm')
		body = await view.render()
		assert 'TestMessage' in body

		view = PromptView(message='TestMessage', size='md')
		body = await view.render()
		assert 'TestMessage' in body

		view = PromptView(message='TestMessage', size='lg')
		body = await view.render()
		assert 'TestMessage' in body



"""
The events manager contains the class that manages custom and abstract callbacks into the system callbacks.
Once a signals is registered here it could be used by string reference. This makes it easy to have dynamically signals
being created by other apps in a single place so it could be used over all apps.

For example you would create your own custom signal if you have a app for your own created game mode script that abstracts
all the raw XML-RPC events into nice structured and maybe even including fetched data from external sources.
"""
import importlib
import logging
import sys

from pyplanet.core.events import Signal


class _SignalManager:
	"""
	Signal Manager class.

	.. note::

		Access this in the app via ``self.context.signals``.

	"""

	def __init__(self):
		self.app_managers = dict()

		self.signals = dict()
		self.callbacks = dict()

		# Reserved signal receivers, this will be filled, and copied to real signals later on.
		self.reserved = dict()
		self.reserved_self = dict()
		#

		self.namespaces = list()

		# This var is used to temporary override namespaces when processing apps.
		self._current_namespace = None

	def register_signal(self, signal, app=None, callback=False):
		"""
		Register a signal to be known in the signalling system.

		:param signal: Signal(s)
		:param app: App context/instance.
		:param callback: Will a callback handle the response (mostly raw callbacks).
		"""
		if isinstance(signal, list):
			for sig in signal:
				self.register_signal(sig)
			return

		if not signal.code:
			raise Exception('Signal code is not valid!')
		if not signal.namespace and self._current_namespace:
			namespace = self._current_namespace
		else:
			namespace = signal.namespace
		code = signal.code

		if not hasattr(signal, 'receivers'):
			instance = signal()
		else:
			instance = signal

		signal_code = '{}:{}'.format(namespace, code)

		if callback:
			self.callbacks[code] = instance
		else:
			self.signals[signal_code] = instance

	def listen(self, signal, target, conditions=None, **kwargs):
		"""
		Register a listing client to the signal given (signal instance or string).

		:param signal: Signal instance or string: "namespace:code"
		:param target: Target method to call.
		:param conditions: Reserved for future purposes.
		"""
		try:
			if not isinstance(signal, Signal):
				signal = self.get_signal(signal)
			signal.register(target, **kwargs)
		except:
			if signal not in self.reserved:
				self.reserved[signal] = list()
			self.reserved[signal].append((target, kwargs))

	def get_callback(self, call_name):
		"""
		Get signal by XML-RPC (script) callback.

		:param call_name: Callback name.
		:return: Signal class or nothing.
		:rtype: pyplanet.core.events.Signal
		"""
		if call_name in self.callbacks:
			return self.callbacks[call_name]
		logging.debug('No callback registered for {}'.format(call_name))
		return None

	def get_signal(self, key):
		"""
		Get signal by key (namespace:code).

		:param key: namespace:code key.
		:return: signal or none
		:rtype: pyplanet.core.events.Signal
		"""
		if key in self.signals:
			return self.signals[key]
		else:
			raise KeyError('No such signal {}!'.format(key))

	def finish_reservations(self):  # pragma: no cover
		"""
		The method will copy all reservations to the actual signals. (PRIVATE)
		"""
		for sig_name, recs in self.reserved.items():
			for func, kwargs in recs:
				try:
					signal = self.get_signal(sig_name)
					signal.connect(func, **kwargs)
				except Exception as e:
					logging.warning('Signal not found: {}, {}'.format(
						sig_name, e
					), exc_info=sys.exc_info())

		for sig_name, recs in self.reserved_self.items():
			for func, slf in recs:
				try:
					signal = self.get_signal(sig_name)
					signal.set_self(func, slf)
				except Exception as e:
					logging.warning(str(e), exc_info=sys.exc_info())

		self.reserved = dict()
		self.reserved_self = dict()

	def init_app(self, app):
		"""
		Initiate app, load all signal/callbacks files. (just import, they should register with decorators).

		:param app: App instance
		:type app: pyplanet.apps.AppConfig
		"""
		self._current_namespace = app.label

		# Import the signals module.
		try:
			importlib.import_module('{}.signals'.format(app.name))
		except ImportError:
			pass
		self._current_namespace = None

		# Import the callbacks module.
		try:
			importlib.import_module('{}.callbacks'.format(app.name))
		except ImportError:
			pass

	async def finish_start(self, *args, **kwargs):
		"""
		Finish startup the core, this will copy reservations. (PRIVATE).
		"""
		self.finish_reservations()

	def create_app_manager(self, app):
		"""
		This method will create the manager instance for the app context.

		:param app: App instance.
		:type app: pyplanet.apps.config.AppConfig
		:return: SignalManager instance for the app.
		:rtype: pyplanet.core.events.manager.AppSignalManager
		"""
		return AppSignalManager(self, app)


class AppSignalManager:

	def __init__(self, manager, app):
		"""
		Create the app manager proxy.
		:param manager: Signal manager (core global).
		:param app: App instance.
		:type manager: pyplanet.core.events.manager._SignalManager
		:type app: pyplanet.apps.config.AppConfig
		"""
		self.manager = manager
		self.app = app

		# Hold these to delete when the context is destroyed.
		self.signals = list()
		self.listeners = list()

	def register_signal(self, signal, callback=False):
		"""
		Register a signal to be known in the signalling system.

		:param signal: Signal(s)
		:param callback: Will a callback handle the response (mostly raw callbacks).
		"""
		self.manager.register_signal(signal, self.app, callback=callback)
		self.signals = list

	def listen(self, signal, target, conditions=None, **kwargs):
		"""
		Register a listing client to the signal given (signal instance or string).

		:param signal: Signal instance or string: "namespace:code"
		:param target: Target method to call.
		:param conditions: Reserved for future purposes.
		"""
		self.manager.listen(signal, target, conditions, **kwargs)
		self.listeners.append((signal, target))

	def get_callback(self, call_name):
		"""
		Get signal by XML-RPC (script) callback.

		:param call_name: Callback name.
		:return: Signal class or nothing.
		:rtype: pyplanet.core.events.Signal
		"""
		return self.manager.get_callback(call_name)

	def get_signal(self, key):
		"""
		Get signal by key (namespace:code).

		:param key: namespace:code key.
		:return: signal or none
		:rtype: pyplanet.core.events.Signal
		"""
		return self.manager.get_signal(key)

	async def on_destroy(self):
		for signal, target in self.listeners:
			try:
				signal.unregister(target)
			except Exception as e:
				logging.exception(e)

		# TODO: Delete signals.


SignalManager = _SignalManager()


def public_signal(cls):
	SignalManager.register_signal(cls)
	return cls


def public_callback(cls):
	SignalManager.register_signal(cls)
	return cls



from pyplanet.core import Controller
from pyplanet.core.events import Callback


async def handle_manialink_answer(source, signal, **kwargs):
	_, player_login, action, raw_values = source
	values = dict()
	if isinstance(raw_values, list):
		for val in raw_values:
			values[val['Name']] = val['Value']
	player = await Controller.instance.player_manager.get_player(login=player_login)

	return dict(player=player, action=action, values=values)


manialink_answer = Callback(
	call='ManiaPlanet.PlayerManialinkPageAnswer',
	namespace='maniaplanet',
	code='manialink_answer',
	target=handle_manialink_answer,
)

"""
:Signal: 
	Player has raised an action on the Manialink.
:Code:
	``maniaplanet:manialink_answer``
:Description:
	Callback sent when a player clicks on an event of a manialink.
:Original Callback:
	`Native` Maniaplanet.PlayerManialinkPageAnswer

:param player: Player instance
:param action: Action name
:param values: Values (in dictionary).
:type player: pyplanet.apps.core.maniaplanet.models.player.Player

.. warning ::

	**Don't use this callback directly, use the abstraction of ``View`` and ``StaticManialink`` to handle events of your
	manialink**!

"""




from pyplanet.views import TemplateView


class StatsView(TemplateView):
	def __init__(self, manager, player, *args, **kwargs):
		self.player = player
		super().__init__(manager, *args, **kwargs)

	def display(self, *args, **kwargs):
		return super().display([self.player.login], **kwargs)



import math


def format_time(time, hide_hours_when_zero=True, hide_milliseconds=False):
	"""
	Format time from integer milliseconds to string format that could be displayed to the end-user.
	
	:param time: Integer time in milliseconds.
	:type time: int
	:param hide_hours_when_zero: Hide the hours when there are zero hours.
	:type hide_hours_when_zero: bool
	:param hide_milliseconds: Hide the milliseconds.
	:type hide_milliseconds: bool
	:return: String output
	:rtype: str
	"""
	hours = math.floor((time / 1000 / 60 / 60))
	minutes = math.floor((time - (hours * 60 * 60 * 1000)) / 1000 / 60)
	seconds = math.floor((time - (hours * 60 * 60 * 1000) - (minutes * 60 * 1000)) / 1000)
	millis = (time - (hours * 60 * 60 * 1000) - (minutes * 60 * 1000) - (seconds * 1000))

	formatted_time = ''
	if hours > 0 or not hide_hours_when_zero:
		formatted_time += '{:02d}:{:02d}:'.format(hours, minutes)
	else:
		formatted_time += '{}:'.format(str(minutes))
	if hide_milliseconds:
		return formatted_time + '{:02d}'.format(seconds)
	return formatted_time + '{:02d}.{:03d}'.format(seconds, millis)



"""
Dev app component.
"""
import json

from pyplanet.contrib.command import Command

from .views.call import CallMenuView


class DevComponent:
	def __init__(self, app):
		"""
		Developer tools component

		:param app: App config instance
		:type app: pyplanet.apps.core.pyplanet.app.PyPlanetConfig
		"""
		self.app = app

	async def on_init(self):
		pass

	async def on_start(self):
		await self.app.instance.permission_manager.register(
			'execute_calls', 'Can execute calls to server.', app=self.app, min_level=3
		)

		await self.app.instance.command_manager.register(
			Command('call', self.admin_call, perms='core.pyplanet:execute_calls', admin=True,
					description='Allows execution of API calls on the dedicated server.')
				.add_param('search', type=str, required=False)
		)

	async def admin_call(self, player, data, **kwargs):
		await self.app.instance.chat('$ff0Please wait... loading methods...', player)

		view = CallMenuView(self.app, player)
		if 'search' in data:
			view.search_text = data.search
		await view.display()
		return



import os
import asynctest

from pyplanet.conf import settings


class TestConfiguration(asynctest.TestCase):

	async def tearDown(self):
		os.environ['PYPLANET_SETTINGS_METHOD'] = 'python'
		settings.reset()

	async def test_lazy_loading(self):
		settings.reset()
		assert settings.configured is False
		_ = settings.DEBUG
		assert settings.configured is True
		assert type(settings.DEBUG) is bool

	async def test_python_backend(self):
		settings.reset()
		os.environ['PYPLANET_SETTINGS_METHOD'] = 'python'

		assert settings.SETTINGS_METHOD == 'python'
		assert type(settings.DEBUG) is bool
		assert settings.configured is True

	async def test_json_backend(self):
		settings.reset()
		os.environ['PYPLANET_SETTINGS_METHOD'] = 'json'

		assert settings.SETTINGS_METHOD == 'json'
		assert type(settings.DEBUG) is bool
		assert settings.configured is True

	async def test_yaml_backend(self):
		settings.reset()
		os.environ['PYPLANET_SETTINGS_METHOD'] = 'yaml'

		assert settings.SETTINGS_METHOD == 'yaml'
		assert type(settings.DEBUG) is bool
		assert settings.configured is True



from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from ..models.local_record import LocalRecord


def upgrade(migrator: SchemaMigrator):
	migrate(
		migrator.add_index(LocalRecord._meta.db_table, [
			'player_id', 'map_id'
		], unique=True),
	)


def downgrade(migrator: SchemaMigrator):
	pass



import functools
import os
import sys

from pyplanet.utils import termcolors


def supports_color():
	"""
	Return True if the running system's terminal supports color,
	and False otherwise.
	"""
	plat = sys.platform
	supported_platform = plat != 'Pocket PC' and (plat != 'win32' or 'ANSICON' in os.environ)

	# isatty is not always implemented, #6223.
	is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
	if not supported_platform or not is_a_tty:
		return False
	return True


class Style:
	pass


def make_style(config_string=''):
	"""
	Create a Style object from the given config_string.

	If config_string is empty pyplanet.utils.termcolors.DEFAULT_PALETTE is used.
	"""

	style = Style()

	color_settings = termcolors.parse_color_setting(config_string)

	# The nocolor palette has all available roles.
	# Use that palette as the basis for populating
	# the palette as defined in the environment.
	for role in termcolors.PALETTES[termcolors.NOCOLOR_PALETTE]:
		if color_settings:
			format = color_settings.get(role, {})
			style_func = termcolors.make_style(**format)
		else:
			def style_func(x):
				return x
		setattr(style, role, style_func)

	# For backwards compatibility,
	# set style for ERROR_OUTPUT == ERROR
	style.ERROR_OUTPUT = style.ERROR

	return style


@functools.lru_cache(maxsize=None)
def no_style():
	"""
	Return a Style object with no color scheme.
	"""
	return make_style('nocolor')


def color_style():
	"""
	Return a Style object from the PyPlanet color scheme.
	"""
	if not supports_color():
		return no_style()
	return make_style(os.environ.get('PYPLANET_COLORS', ''))



from importlib import import_module

from pyplanet.core.management import CommandError
from pyplanet.core.management.templates import TemplateCommand


class Command(TemplateCommand):
	help = 'Create a new app in your local apps. package. You should consider moving it up once you want to publish it!'

	requires_system_checks = False
	requires_migrations_checks = False

	template_files = [
		'__init__.py',
		'models.py',
	]

	def handle(self, **options):
		app_name, target = options.pop('name'), options.pop('directory')
		self.validate_name(app_name, 'app')

		# Check that the app_name cannot be imported.
		try:
			import_module(app_name)
		except ImportError:
			pass
		else:
			raise CommandError(
				"%r conflicts with the name of an existing Python module and "
				"cannot be used as an app name. Please try another name." % app_name
			)

		super().handle('app', app_name, target, **options)



import asyncio

from pyplanet.apps import AppConfig
from pyplanet.contrib import CoreContrib
from pyplanet.contrib.setting.core_settings import performance_mode
from pyplanet.contrib.setting.exceptions import SettingException


class _BaseSettingManager:
	def __init__(self, instance):
		"""
		Initiate, should only be done from the core instance.

		:param instance: Instance.
		:type instance: pyplanet.core.instance.Instance
		"""
		self._instance = instance
		self._settings = list()
		self._app = None

	async def register(self, *settings):
		"""
		Register your setting(s). This will create default values when the setting has not yet been inited before.

		:param settings: Setting(s) given.
		:type settings: pyplanet.contrib.setting.setting._Setting
		"""
		# Check if setting has a value, then fetch it, if not, create new entry with default or none.
		await asyncio.gather(*[s.initiate_setting() for s in settings])

		# Register the setting.
		self._settings.extend(settings)


class GlobalSettingManager(_BaseSettingManager, CoreContrib):
	"""
	Global Setting manager is available at the instance. ``instance.setting_manager``.

	.. warning::

		Don't use the setting_manager for registering app settings! Use the app setting manager instead!

		Don't initiate this class yourself.

	"""

	def __init__(self, instance):
		super().__init__(instance)
		self.app_managers = dict()

	async def on_start(self):
		# Register core global settings.
		await self.register(performance_mode)

	def create_app_manager(self, app_config):
		"""
		Create app setting manager.

		:param app_config: App Config instance.
		:type app_config: pyplanet.apps.config.AppConfig
		:return: Setting Manager
		:rtype: pyplanet.contrib.setting.manager.AppSettingManager
		"""
		if app_config.label not in self.app_managers:
			self.app_managers[app_config.label] = AppSettingManager(self._instance, app_config)
		return self.app_managers[app_config.label]

	def get_app_manager(self, app):
		"""
		Get the app manager for a specified app label or config instance.

		:param app: App label in string or the app config instance.
		:return: App manager instance.
		:rtype: pyplanet.contrib.setting.manager.AppSettingManager
		"""
		if isinstance(app, AppConfig):
			app = app.label
		return self.app_managers[app]

	@property
	def recursive_settings(self):
		"""
		Retrieve all settings, of all submanagers.
		"""
		for setting in self._settings:
			yield setting
		for app, manager in self.app_managers.items():
			for setting in manager._settings:
				yield setting

	async def get_setting(self, app_label, key, prefetch_values=True):
		"""
		Get setting by key and optionally fetch the value if not yet fetched.

		:param app_label: Namespace (mostly app label).
		:param key: Key string
		:param prefetch_values: Prefetch the values if not yet fetched?
		:return: Setting instance.
		:raise: SettingException
		"""
		if app_label is None:
			setting = None
			for s in self._settings:
				if s.key == key:
					setting = s
					break

			if not setting:
				raise SettingException('Setting with key not found')

			if prefetch_values and setting._value[0] is False:
				await setting.get_value()
			return setting
		return await self.get_app_manager(app_label).get_setting(key, prefetch_values)

	async def get_apps(self, prefetch_values=True):
		"""
		Get all the app label + names for all the settings we can find in our registry.
		Returns a dict with label as key, and count + name as values.

		:param prefetch_values: Prefetch the values in this call. Defaults to True.
		:return: List with setting objects.
		"""
		apps = dict()
		if prefetch_values:
			await asyncio.gather(*[
				s.get_value(refresh=True) for s in self.recursive_settings
			])

		for setting in self.recursive_settings:
			if setting.app_label not in apps:
				apps[setting.app_label] = dict(
					count=0,
					name=self._instance.apps.apps[setting.app_label].name,
					app=self._instance.apps.apps[setting.app_label],
					settings=list()
				)
			apps[setting.app_label]['count'] += 1
			apps[setting.app_label]['settings'].append(setting)
		return apps

	async def get_categories(self, prefetch_values=True):
		"""
		Get all the categories we have registered.
		Returns a dict with label as key, and count + name as values.

		:param prefetch_values: Prefetch the values in this call. Defaults to True.
		:return: List with setting objects.
		"""
		cats = dict()
		if prefetch_values:
			await asyncio.gather(*[
				s.get_value(refresh=True) for s in self.recursive_settings
			])

		for setting in self.recursive_settings:
			if setting.category not in cats:
				cats[setting.category] = dict(
					count=0,
					name=setting.category,
					settings=list()
				)
			cats[setting.category]['count'] += 1
			cats[setting.category]['settings'].append(setting)
		return cats

	async def get_all(self, prefetch_values=True):
		"""
		Retrieve a list of settings, with prefetched values, so get_value is almost instant (or use ._value, not recommended).

		:param prefetch_values: Prefetch the values in this call. Defaults to True.
		:return: List with setting objects.
		"""
		if prefetch_values:
			await asyncio.gather(*[
				s.get_value(refresh=True) for s in self.recursive_settings
			])
		return self.recursive_settings


class AppSettingManager(_BaseSettingManager):
	"""
	The local app setting manager is the one you should use to register settings to inside of your app.

	You can use this manager like this:

	.. code-block:: python

		from pyplanet.contrib.setting import Setting

		async def on_start(self):
			await self.context.setting.register(
				Setting('feature_a', 'Enable feature A', Setting.CAT_FEATURES, type=bool, description='Enable feature A'),
				Setting('feature_b', 'Enable feature B', Setting.CAT_FEATURES, type=bool, description='Enable feature B'),
			)

	For more information about the settings, categories, types, and all other options. Look at the ``Settings``
	documentation.

	.. warning::

		Don't initiate this class yourself.

	"""

	def __init__(self, instance, app):
		"""
		Initiate app setting manager.

		:param instance: Controller instance.
		:param app: App Config instance.
		:type instance: pyplanet.core.instance.Instance
		:type app: pyplanet.apps.config.AppConfig
		"""
		super().__init__(instance)
		self._app = app

	async def register(self, *settings):
		"""
		Register your setting(s). This will create default values when the setting has not yet been inited before.

		:param settings: Setting(s) given.
		:type settings: pyplanet.contrib.setting.setting._Setting
		"""
		# Set app label on all setting objects.
		for setting in settings:
			setting.app_label = self._app.label

		# Check if setting has a value, then fetch it, if not, create new entry with default or none.
		await asyncio.gather(*[s.initiate_setting() for s in settings])

		# Register the setting.
		self._settings.extend(settings)

	async def get_setting(self, key, prefetch_values=True):
		"""
		Get setting by key and optionally fetch the value if not yet fetched.

		:param key: Key string
		:param prefetch_values: Prefetch the values if not yet fetched?
		:return: Setting instance.
		:raise: SettingException
		"""
		setting = None
		for s in self._settings:
			if s.key == key:
				setting = s
				break

		if not setting:
			raise SettingException('Setting with key not found')

		if prefetch_values and setting._value[0] is False:
			await setting.get_value()
		return setting

	def get_categories(self):
		"""
		Get all the categories we have registered.
		Returns a dict with label as key, and count + name as values.
		"""
		cats = dict()
		for setting in self._settings:
			if setting.category not in cats:
				cats[setting.category] = dict(count=0)
			cats[setting.category]['count'] += 1
		return cats

	async def get_all(self, prefetch_values=True):
		"""
		Retrieve a list of settings, with prefetched values, so get_value is almost instant (or use ._value, not recommended).

		:param prefetch_values: Prefetch the values in this call. Defaults to True.
		:return: List with setting objects.
		"""
		if prefetch_values:
			await asyncio.gather(*[
				s.get_value(refresh=True) for s in self._settings
			])
		return self._settings



import pymysql
import pymysql.cursors


class BaseConverter:
	"""
	Base Converter is the abstract converter class.

	Please take a look at the other classes bellow.
	"""
	def __init__(
		self, instance, db_type, db_host, db_name, db_user=None, db_password=None, db_port=None, prefix=None,
		charset='utf8'
	):
		"""
		Create converter.

		:param instance: Controller instance.
		:param db_type: Type, mysql by default.
		:param db_host: Hostname
		:param db_name: Name of db schema/database
		:param db_user: Username
		:param db_password: Password
		:param db_port: Port.
		:param prefix: Table prefix.
		:param charset: Charset of source db. Only supporting utf8 now.
		:param extra: Any extra parameters given.
		:type instance: pyplanet.core.instance.Instance
		"""
		self.instance = instance

		self.db_type = db_type
		self.db_host = db_host
		self.db_name = db_name
		self.db_user = db_user
		self.db_password = db_password
		self.db_port = db_port
		self.prefix = prefix or ''
		self.charset = charset

		self.connection = None

	async def connect(self):
		await self.instance.db.connect()
		await self.instance.apps.discover()
		await self.instance.db.initiate()

		if self.db_type != 'mysql':
			raise Exception('We only support mysql converting right now!')

		self.connection = pymysql.connect(
			host=self.db_host, user=self.db_user, password=self.db_password, db=self.db_name, charset=self.charset,
			port=self.db_port or 3306,
			cursorclass=pymysql.cursors.DictCursor
		)

	async def start(self):
		if not self.connection:
			raise Exception('Please connect first (connect()).')
		return await self.migrate(self.connection)

	async def migrate(self, source_connection):
		raise NotImplementedError



import asyncssh
import logging
import os
import async_generator
import asyncio_extras

from pyplanet.core.storage import StorageDriver

logger = logging.getLogger(__name__)


class SFTPDriver(StorageDriver):
	"""
	SFTP storage driver is using the asyncssh module to access storage that is situated remotely.
	
	.. warning::
	
		This driver is not ready for production use!!
	
	:option HOST: Hostname of destinotion server.
	:option PORT: Port destinotion server.
	:option USERNAME: Username of the user account.
	:option PASSWORD: Password of the user account. (optional if you use public/private keys).
	:option KNOWN_HOSTS: File to the Known Hosts file.
	:option CLIENT_KEYS: Array with client private keys.
	:option PASSPHRASE: Passphrase to unlock private key(s).
	:option KWARGS: Any other options that will be passed to ``asyncssh``.
	"""

	def __init__(self, instance, config: dict = None):
		super().__init__(instance, config)

		# Extract config to local vars.
		self.host = config['HOST']
		self.port = int(config['PORT']) if 'PORT' in config else 22
		self.username = config['USERNAME']
		self.password = config['PASSWORD'] if 'PASSWORD' in config else None
		self.known_hosts = config['KNOWN_HOSTS'] if 'KNOWN_HOSTS' in config and isinstance(config['KNOWN_HOSTS'], list) else []
		self.client_keys = config['CLIENT_KEYS'] if 'CLIENT_KEYS' in config and isinstance(config['CLIENT_KEYS'], list) else []
		self.passphrase = config['PASSPHRASE'] if 'PASSPHRASE' in config else None
		self.kwargs = config['KWARGS'] if 'KWARGS' in config and isinstance(config['KWARGS'], dict) else dict()

		self.options = dict(
			host=self.host, port=self.port, known_hosts=self.known_hosts, username=self.username, password=self.password,
			client_keys=self.client_keys, passphrase=self.passphrase,
		)

	@asyncio_extras.async_contextmanager
	async def connect(self):
		ssh = await asyncssh.connect(
			host=self.host, port=self.port, known_hosts=self.known_hosts, username=self.username, password=self.password,
			client_keys=self.client_keys, passphrase=self.passphrase,
		).__aenter__()
		await async_generator.yield_(ssh)
		await ssh.__aexit__()

	@asyncio_extras.async_contextmanager
	async def connect_sftp(self):
		"""
		Get sftp client.
		
		:return: Sftp client.
		:rtype: asyncssh.SFTPClient
		"""
		ssh = await self.connect().__aenter__()
		sftp = await ssh.start_sftp_client().__aenter__()
		await async_generator.yield_(sftp)
		await sftp.__aexit__()
		await ssh.__aexit__()

	async def chmod(self, path: str, mode: int, **kwargs):
		async with self.connect_sftp() as sftp:
			await sftp.chmod(self.absolute(path), mode)

	async def chown(self, path: str, uid: int, gid: int, **kwargs):
		async with self.connect_sftp() as sftp:
			await sftp.chown(self.absolute(path), uid, gid)

	async def close(self, **kwargs):
		pass

	@asyncio_extras.async_contextmanager
	async def open(self, filename: str, mode: str = 'r', **kwargs):
		sftp = await self.connect_sftp().__aenter__()
		fh = await sftp.open(self.absolute(filename), mode, **kwargs).__aenter__()
		await async_generator.yield_(fh)
		await fh.__aexit__()
		await sftp.__aexit__()

	async def get(self, remotepath: str, localpath: str, **kwargs):
		async with self.connect_sftp() as sftp:
			return await sftp.get(self.absolute(remotepath), localpath, preserve=True, follow_symlinks=True, **kwargs)

	async def put(self, localpath: str, remotepath: str, **kwargs):
		async with self.connect_sftp() as sftp:
			await sftp.put(localpath, self.absolute(remotepath), preserve=True, follow_symlinks=True, **kwargs)

	async def listdir(self, path='.', **kwargs):
		async with self.connect_sftp() as sftp:
			return await sftp.listdir(self.absolute(path))

	async def mkdir(self, path, mode=511, **kwargs):
		async with self.connect_sftp() as sftp:
			attrs = asyncssh.SFTPAttrs()
			attrs.permissions = mode
			for k, v in kwargs.items():
				attrs.__setattr__(k, v)
			await sftp.mkdir(self.absolute(path), attrs)

	async def remove(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			await sftp.remove(self.absolute(path))

	async def rename(self, oldpath: str, newpath: str, **kwargs):
		async with self.connect_sftp() as sftp:
			await sftp.rename(self.absolute(oldpath), self.absolute(newpath))

	async def rmdir(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			async def rm(file):
				if await sftp.isdir(file):
					files = await sftp.listdir(os.path.join(self.absolute(path), file))
					for subfile in files:
						await rm(subfile)
				else:
					await sftp.remove(file)
			await rm(self.absolute(path))

	async def stat(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			return await sftp.stat(self.absolute(path))

	async def symlink(self, source: str, dest: str, **kwargs):
		async with self.connect_sftp() as sftp:
			await sftp.symlink(self.absolute(source), self.absolute(dest))

	async def exists(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			return await sftp.exists(self.absolute(path))

	async def is_file(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			return await sftp.isfile(self.absolute(path))

	async def is_dir(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			return await sftp.isdir(self.absolute(path))

	async def is_link(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			return await sftp.islink(self.absolute(path))

	async def touch(self, path: str, **kwargs):
		async with self.connect_sftp() as sftp:
			async with sftp.open(self.absolute(path), 'w+') as fh:
				await fh.write('')

	def openable(self):
		return True



from argparse import Namespace

from pyplanet.contrib.command.exceptions import (
	ParamValidateException, ParamParseException, ParamException,
	NotValidated,
	InvalidParamException)


class ParameterParser:
	"""
	Parameter Parser.
	
	.. todo::
	
		Write introduction + examples.
		
	"""

	def __init__(self, prog=None):
		self.prog = prog
		self.params = list()

		self._errors = list()
		self.data = object()

	def add_param(
		self, name: str,
		nargs=1,
		type=str,
		default=None,
		required: bool=True,
		help: str=None,
		dest: str=None,
	):
		"""
		Add positional parameter.
		
		:param name: Name of parameter, will be used to store result into!
		:param nargs: Number of arguments, use integer or '*' for multiple or infinite.
		:param type: Type of value, keep str to match all types. Use any other to try to parse to the type.
		:param default: Default value when no value is given.
		:param required: Set the parameter required state, defaults to true.
		:param help: Help text to display when parameter is invalid or not given and required.
		:param dest: Destination to save into namespace result (defaults to name).
		:return: parser instance
		:rtype: pyplanet.contrib.command.ParameterParser
		"""
		self.params.append(dict(
			name=name, nargs=nargs, type=type, default=default, required=required, help=help, dest=dest
		))
		return self

	def parse_parameter(self, param, input, position):
		"""
		Validate and parse param value at input position.
		
		:param param: Param dict given.
		:param input: Full params input (without command or namespace!)
		:param position: Current seek position.
		:type param: dict
		:type input: list
		:type position: int
		:return: value.
		"""
		try:
			part = input[position]
		except IndexError:
			if param['required'] is False:
				return param['default']
			else:
				raise ParamValidateException('param \'{}\' is required'.format(param['name']))

		value = None
		# If we have multiple arguments of the same type, parse it internally.
		if isinstance(param['nargs'], int) and param['nargs'] > 1:
			# We need to clone to prevent infinite loop
			nparam = dict()
			nparam.update(param)
			nparam['nargs'] = 1

			value = []
			errors = []
			for i in range(1, param['nargs']):
				try:
					value.append(self.parse_parameter(nparam, input, position + i))
				except ParamException as e:
					errors.append(str(e))
			if len(errors) > 0:
				raise ParamParseException(', '.join(errors))

		# If we expect multiple (infinite) occurrences.
		elif isinstance(param['nargs'], str) and param['nargs'] == '*':
			# We need to clone to prevent infinite loop
			nparam = dict()
			nparam.update(param)
			nparam['nargs'] = 1

			if value:
				value = [value]
			else:
				value = []

			for i in range(0, len(input)):
				try:
					extra_value = self.parse_parameter(nparam, input, position + i)
					if extra_value is not None:
						value.append(extra_value)
				except ParamException:
					# We will stop here.
					break

		else:
			if param['type'] is int:
				try:
					value = int(part)
				except ValueError:
					raise ParamParseException('param \'{}\' must be an integer'.format(param['name']))
			elif param['type'] is str:
				value = part
			else:
				raise InvalidParamException('Type of parameter \'{}\' is not known.'.format(param['name']))

		return value

	def parse(self, argv):
		"""
		Parse arguments.
		
		:param argv: arguments.
		"""
		values = dict()

		self.data = None
		self._errors = list()

		for idx, param in enumerate(self.params):
			try:
				values[param['dest'] or param['name']] = self.parse_parameter(param, argv, idx)
			except ParamException as e:
				self._errors.append(str(e))

		self.data = Namespace(**values)

	def is_valid(self):
		"""
		Is data valid?
		
		:return: boolean
		"""
		if self.data is None:
			raise NotValidated('Parameters not yet parsed, call parse() first.')
		return len(self._errors) == 0

	@property
	def errors(self):
		"""
		Get errors.
		
		:return: array of strings.
		:rtype: list
		"""
		return self._errors



import os
import yaml

from pyplanet.conf.backends.file import FileConfigBackend
from pyplanet.core.exceptions import ImproperlyConfigured


class YamlConfigBackend(FileConfigBackend):
	name = 'yaml'
	files = ['base.yaml', 'apps.yaml']
	required_files = ['base.yaml', 'apps.yaml']

	def load(self):
		# Prepare + load directory.
		super().load()

		# Load the files and parse Yaml.
		parsed_settings = dict()

		try:
			for file_name in self.files:
				file_path = os.path.join(self.directory, file_name)
				with open(file_path, 'r') as file_handle:
					parsed_settings.update(yaml.safe_load(file_handle))
		except (yaml.YAMLError, yaml.MarkedYAMLError) as e:
			raise ImproperlyConfigured(
				'Your settings file(s) contain invalid YAML syntax! Please fix and restart!, {}'.format(str(e))
			)

		# Loop and set in local settings (+ uppercase keys).
		for key, value in parsed_settings.items():
			self.settings[key.upper()] = value



import random

import aiohttp
import logging
import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.ads.view import PayPalLogoView, DiscordLogoView
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.setting import Setting
from pyplanet.contrib.command import Command
from pyplanet import __version__ as pyplanet_version


class Ads(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		# Initiate the logo views.
		self.paypal_view = PayPalLogoView(self, manager=self.context.ui)
		self.discord_view = DiscordLogoView(self, manager=self.context.ui)

		self.paypal_donation_url = None
		self.discord_join_url = None
		self.discord_server_id = None

		self.random_future = None
		self.random_last_message = None

		# Initiate settings.
		self.setting_enable_paypal = Setting(
			'paypal_enable', 'Show PayPal Donation Logo', Setting.CAT_DESIGN, type=bool,
			description='Show the PayPal logo with clickable link to donation page', change_target=self.reload_settings,
			default=False
		)
		self.setting_paypal_url = Setting(
			'paypal_url', 'PayPal Donation URL', Setting.CAT_KEYS, type=str,
			description='Your personalized PayPal donation link.', change_target=self.reload_settings,
		)

		self.setting_enable_discord = Setting(
			'discord_enable', 'Show Discord Logo', Setting.CAT_DESIGN, type=bool,
			description='Show the Discord logo with clickable link to join the server', change_target=self.reload_settings,
			default=False
		)
		self.setting_discord_join_url = Setting(
			'discord_url', 'Discord Invite URL', Setting.CAT_KEYS, type=str,
			description='Set the link to not expire in discord for this to keep working!',
			default='https://discord.gg/x9amg5K', change_target=self.reload_settings
		)
		self.setting_discord_server_id = Setting(
			'discord_id', 'Discord Server ID', Setting.CAT_KEYS, type=str,
			description='Get your discord server ID and enable the widget in the discord server settings!',
			default='158349559959912448', change_target=self.reload_settings
		)

		self.setting_random_messages = Setting(
			'random_messages', 'Random Messages', Setting.CAT_BEHAVIOUR, type=list,
			description='Show random messages at random moments in the chat. '
						'The messages will be displayed every time the interval passes (another setting).',
			default=None, change_target=self.reload_settings
		)
		self.setting_random_messages_interval = Setting(
			'random_messages_interval', 'Random Messages Interval', Setting.CAT_BEHAVIOUR, type=int,
			description='The interval of the random messages in seconds. Default 2 minutes (120 seconds)',
			default=120, change_target=self.reload_settings
		)

	async def on_start(self):
		# Register signals.
		self.context.signals.listen(mp_signals.player.player_connect, self.on_connect)

		# Register commands.
		await self.instance.command_manager.register(
			Command(command='paypal', target=self.chat_paypal, admin=False)
		)
		await self.instance.command_manager.register(
			Command(command='discord', target=self.chat_discord, admin=False),
		)

		# Register settings.
		await self.context.setting.register(
			self.setting_enable_discord,
			self.setting_enable_paypal,

			self.setting_paypal_url,
			self.setting_discord_join_url,
			self.setting_discord_server_id,

			self.setting_random_messages,
			self.setting_random_messages_interval,
		)

		# Force reload of settings.
		await self.reload_settings()

		# Random Messages Loop.
		self.random_future = asyncio.ensure_future(self.random_messages_loop())

		# Call the display method.
		await self.display()

	async def display(self, logins=None):
		show_paypal = await self.setting_enable_paypal.get_value()
		show_discord = await self.setting_enable_discord.get_value()

		if show_paypal:
			await self.paypal_view.display(player_logins=logins)
		if show_discord:
			await self.discord_view.display(player_logins=logins)

	async def hide_all(self):
		await asyncio.gather(
			self.paypal_view.hide(),
			self.discord_view.hide(),
		)

	async def on_connect(self, player, *args, **kwargs):
		await self.display(logins=[player.login])

	async def reload_settings(self, *args, **kwargs):
		# Verify PayPal settings and get donation url.
		if await self.setting_enable_paypal.get_value():
			url = await self.setting_paypal_url.get_value()
			if not url:
				await self.instance.chat('$f00$iInvalid PayPal Donation URL, none given!')
			else:
				with aiohttp.ClientSession(headers={'User-Agent': 'PyPlanet/{}'.format(pyplanet_version)}) as session:
					try:
						async with session.get(url) as _:
							pass
					except Exception as e:
						await self.instance.chat('$f00$iInvalid PayPal Donation URL')
						logging.error('Error with checking PayPal donation URL.')
						logging.error(e)
				self.paypal_donation_url = url

		# Verify discord settings.
		if await self.setting_enable_discord.get_value():
			url_setting = await self.setting_discord_join_url.get_value(refresh=True)
			id_setting = await self.setting_discord_server_id.get_value(refresh=True)
			url_validity = await self.is_discord_url_valid(url_setting)

			if url_validity:
				self.discord_join_url = url_setting
			else:
				await self.instance.chat('$f00$iInvalid Discord Invite URL')

			id_validity = await self.is_discord_url_valid("https://discordapp.com/api/guilds/" + id_setting + "/widget.json")
			if id_validity:
				self.discord_server_id = id_setting
			else:
				await self.instance.chat('$f00$iInvalid Discord Server ID')

		# Hide & display widgets.
		await self.hide_all()
		await self.display()

	async def random_messages_loop(self):
		while True:
			sleep_interval = abs(await self.setting_random_messages_interval.get_value())
			if sleep_interval == 0:
				sleep_interval = 1

			logging.getLogger(__name__).debug('Random Messages, sleeping for {} seconds'.format(sleep_interval))
			await asyncio.sleep(sleep_interval)

			# Ignore when empty
			if not await self.setting_random_messages.get_value():
				continue

			try:
				await self.send_random_message()
			except Exception as e:
				logging.getLogger(__name__).error('Error sending random message: ', str(e))

	async def send_random_message(self):
		messages = await self.setting_random_messages.get_value()
		if not messages:
			return

		unique = False
		message = ''
		while not unique:
			message = random.choice(messages)
			unique = len(messages) == 1 or not self.random_last_message or self.random_last_message != message

		self.random_last_message = message

		if not message.startswith('$'):
			message = '$06f' + message
		await self.instance.chat(message)

	async def chat_paypal(self, player, *args, **kwargs):
		if await self.setting_enable_paypal.get_value():
			message = "$ff0$iYou can donate to us $l[{}]here!".format(self.paypal_donation_url)
			await self.instance.chat(message, player)

	async def chat_discord(self, player, *args, **kwargs):
		if await self.setting_enable_discord.get_value():
			users = await self.get_discord_users()
			if users:
				join_url_link = '$l[' + self.discord_join_url + ']Join our Discord$l! '
				message = '$ff0$i{}There are currently {} users and {} bots online.' \
					.format(join_url_link, users[0], users[1])
				await self.instance.chat(message, player)

	async def get_discord_users(self):
		url = "https://discordapp.com/api/guilds/" + self.discord_server_id + "/widget.json"

		with aiohttp.ClientSession(headers={'User-Agent': 'PyPlanet/{}'.format(pyplanet_version)}) as session:
			try:
				async with session.get(url) as response:
					data = await response.json()
					non_bot_users = []
					bots = []
					for i in data['members']:
						if 'bot' in i:
							bots.append(i)
						else:
							non_bot_users.append(i)
					online_users = len(non_bot_users)
					return [int(online_users), int(len(bots))]
			except Exception as e:
				logging.error('Error with retrieving Discord user list')
				logging.error(e)
				return False

	async def is_discord_url_valid(self, url):
		with aiohttp.ClientSession(headers={'User-Agent': 'PyPlanet/{}'.format(pyplanet_version)}) as session:
			try:
				async with session.get(url) as _:
					return True
			except Exception as e:
				logging.error('Error with checking for valid Discord URL')
				logging.error(e)
				return False



import asynctest

from pyplanet.core.instance import Controller


class TestGbx(asynctest.TestCase):
	async def test_gbx_init(self):
		instance = Controller.prepare(name='default').instance
		await instance.gbx.connect()
		self.assertGreater(len(instance.gbx.gbx_methods), 0)
		await instance.gbx.disconnect()
		del instance

	async def test_gbx_game_infos(self):
		instance = Controller.prepare(name='default').instance
		await instance.gbx.connect()

		assert instance.game.server_player_login is not None
		assert instance.game.dedicated_build is not None
		assert instance.game.dedicated_version is not None
		assert instance.game.dedicated_api_version is not None
		assert instance.game.dedicated_title is not None

		assert instance.game.server_is_dedicated is True
		assert instance.game.server_is_server is True
		assert instance.game.server_ip is not None
		assert instance.game.server_p2p_port is not None
		assert instance.game.server_port is not None
		assert instance.game.server_player_login
		assert instance.game.server_player_id is not None
		assert instance.game.server_download_rate is not None
		assert instance.game.server_upload_rate is not None
		assert instance.game.server_data_dir is not None
		assert instance.game.server_map_dir is not None
		assert instance.game.server_skin_dir is not None
		assert instance.game.game == 'tm' or instance.game.game == 'sm'

		await instance.gbx.disconnect()
		del instance

	async def test_gbx_query(self):
		instance = Controller.prepare(name='default').instance
		await instance.gbx.connect()

		# Direct way.
		version_info = await instance.gbx.execute('GetVersion')
		assert isinstance(version_info, dict)
		assert len(version_info.keys()) > 0

		# Prepared way.
		query = instance.gbx.prepare('GetVersion')
		version_info = await query.execute()
		assert isinstance(version_info, dict)
		assert len(version_info.keys()) > 0

		await instance.gbx.disconnect()
		del instance

	async def test_script_query(self):
		instance = Controller.prepare(name='default').instance
		await instance.gbx.connect()

		# Direct way.
		pause_info = await instance.gbx.script('Maniaplanet.Pause.GetStatus')
		assert isinstance(pause_info, dict)
		assert len(pause_info.keys()) > 0

		# Prepared way.
		query = instance.gbx.prepare('Maniaplanet.Pause.GetStatus')
		pause_info = await query.execute()
		assert isinstance(pause_info, dict)
		assert len(pause_info.keys()) > 0

		await instance.gbx.disconnect()
		del instance

	async def test_multicall(self):
		instance = Controller.prepare(name='default').instance
		await instance.gbx.connect()

		version_info, ladder_info, pause_info = await instance.gbx.multicall(
			instance.gbx.prepare('GetVersion'),
			instance.gbx.prepare('GetLadderServerLimits'),
			instance.gbx.prepare('Maniaplanet.Pause.GetStatus'),  # WARNING: Scripted can't be combined with normal for now.
		)
		assert isinstance(version_info, dict)
		assert len(version_info.keys()) > 0
		assert isinstance(ladder_info, dict)
		assert len(ladder_info.keys()) > 0
		# BECAUSE WE CAN'T COMBINE INSIDE OF MULTICALL. BECAUSE SCRIPTED IS RETURNED AS SCRIPT-CALLBACKS!
		assert pause_info is True

		await instance.gbx.disconnect()
		del instance



"""
This file contains the local settings and overrides the default ones.
"""
import os

# Set the root_path to the root of our project.
ROOT_PATH = os.path.dirname(os.path.dirname(__file__))

# Set the temporary location for the project.
TMP_PATH = os.path.join(ROOT_PATH, 'tmp')

# Create temporary folder if not exists.
if not os.path.exists(TMP_PATH):
	os.mkdir(TMP_PATH)

# Enable debug mode to get verbose output, not report any errors and dynamically use the DEBUG in your code
# for extra verbosity of logging/output.
DEBUG = True # bool(os.environ.get('PYPLANET_DEBUG', False))

# Add your pools (the controller instances per dedicated here) or leave as it is to use a single instance only.
POOLS = [
	'default',
]

# Owners are logins of the server owners, the owners always get *ALL* the permissions in the system.
OWNERS = {
	'default': [
		'your-maniaplanet-login'
	]
}

# Databases configuration holds an dictionary with information of the database backend.
# Please refer to the documentation for all examples.
MYSQL = {
	'ENGINE': 'peewee_async.MySQLDatabase',
	'NAME': 'pyplanet',
	'OPTIONS': {
		'host': 'localhost',
		'user': 'root',
		'password': 'Password12!',
		'charset': 'utf8mb4',
	}
}
POSTGRESQL = {
	'ENGINE': 'peewee_async.PostgresqlDatabase',
	'NAME': 'pyplanet',
	'OPTIONS': {
		'host': 'localhost',
		'user': 'pyplanet',
		'password': 'pyplanet',
		'autocommit': True,
	}
}
DATABASE_ENGINE = None
TOX_ENV = os.getenv('TOXENV', 'py36-unit-mysql')
if 'mysql' in TOX_ENV:
	DATABASE_ENGINE = MYSQL
elif 'postgresql' in TOX_ENV:
	DATABASE_ENGINE = POSTGRESQL
DATABASES = {
	'default': DATABASE_ENGINE
}

# Dedicated configuration holds the different dedicated servers that the instances will run on including the names of
# the instances.
DEDICATED = {
	'default': {
		'HOST': '127.0.0.1',
		'PORT': '5000',
		'USER': 'SuperAdmin',
		'PASSWORD': 'SuperAdmin',
	}
}

# The storage configuration contains the same instance mapping of the dedicated servers and is used
# to access the filesystem on the dedicated server location.
# Please refer to the documentation for more information.
STORAGE = {
	'default': {
		'DRIVER': 'pyplanet.core.storage.drivers.local.LocalDriver',
		'OPTIONS': {},
	}
}

MAP_MATCHSETTINGS = 'test.txt'



"""
The MX API client class.
"""
import asyncio
import logging
import aiohttp
import re

from pyplanet import __version__ as pyplanet_version
from pyplanet.apps.contrib.mx.exceptions import MXMapNotFound, MXInvalidResponse

logger = logging.getLogger(__name__)


class MXApi:
	def __init__(self, server_login=None):
		self.server_login = server_login
		self.cookie_jar = aiohttp.CookieJar()
		self.session = None
		self.site = None
		self.key = None
		self.map_info_page_size = 1

	def base_url(self, api=False):
		if self.site == 'tm':
			if api:
				return 'https://{site}.mania.exchange/api'.format(site=self.site)
			return 'https://{site}.mania.exchange'.format(site=self.site)
		elif self.site == 'tmnext':
			if api:
				return 'https://trackmania.exchange/api'
			return 'https://trackmania.exchange'
		elif self.site == 'sm':	
			if api:
				return 'https://api.mania-exchange.com/sm'
			return 'https://sm.mania-exchange.com'

	async def create_session(self):
		self.session = await aiohttp.ClientSession(
			cookie_jar=self.cookie_jar,
			headers={
				'User-Agent': 'PyPlanet/{}'.format(pyplanet_version),
				'X-ManiaPlanet-ServerLogin': self.server_login
			}
		).__aenter__()

	async def close_session(self):
		if self.session and hasattr(self.session, '__aexit__'):
			await self.session.__aexit__()
	
	async def mx_random(self):
		# Regular Expression to extract the MX-ID from a /tracksearch2/random/.
		mx_pattern = r'\d+'
		mx_id_regex = re.compile(mx_pattern)
		url = '{}/tracksearch2/random'.format(self.base_url())
		response = await self.session.get(url)
		text = str(response.url)
		matches = re.search(mx_id_regex, text)
		if not matches:
			return None
		return str(matches.group(0))
	
	async def search(self, options, **kwargs):
		if options is None:
			options = {
				"api": "on",
				"mode": 0,
				"style": 0,
				"order": -1,
				"length": -1,
				"page": 0,
				"gv": 1,
				"limit": 150
			}

		if self.key:
			options['key'] = self.key

		url = '{}/tracksearch2/search'.format(self.base_url())
		response = await self.session.get(url, params=options)

		if response.status == 404:
			raise MXMapNotFound('Got not found status from ManiaExchange: {}'.format(response.status))
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))

		maps = list()
		json = await response.json()
		for info in json['results']:
			# Parse some differences between the api game endpoints.
			mx_id = info['TrackID'] if 'TrackID' in info else info['MapID']
			info['MapID'] = mx_id
			info['MapUID'] = info['TrackUID'] if 'TrackUID' in info else info['MapUID']
			maps.append(info)
		return maps

	async def search_pack(self, options, **kwargs):
		if options is None:
			options = dict()

		if self.key:
			options['key'] = self.key

		options['api'] = 'on'

		url = '{}/mappacksearch/search'.format(self.base_url())
		response = await self.session.get(url, params=options)

		if response.status == 404:
			raise MXMapNotFound('Got not found status from ManiaExchange: {}'.format(response.status))
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))

		maps = list()
		json = await response.json()
		if not 'results' in json:
			return list()

		for info in json['results']:
			# Parse some differences between the api game endpoints.
			maps.append(info)

		return maps

	async def map_info(self, *ids):
		if not len(ids):
			return list()
		if isinstance(ids, str) or isinstance(ids, int):
			# In case just one value is being passed, put it into an array.
			ids = [ids]

		# Split the map identifiers into groups, as the ManiaExchange API only accepts a limited amount of maps in one request.
		split_map_ids = [ids[i * self.map_info_page_size:(i + 1) * self.map_info_page_size] for i in range((len(ids) + self.map_info_page_size - 1) // self.map_info_page_size)]
		split_results = list()
		coros = list()
		for split_ids in split_map_ids:
			coros.append(self.map_info_page(split_ids))
		split_results = await asyncio.gather(*coros)

		# Join the multiple result lists back into one list.
		return [map for map_list in split_results for map in map_list]
	
	async def map_offline_record(self, trackid):
		
		url = '{base}/replays/get_replays/{id}/1'.format(base=self.base_url(True), id=trackid)
		params = {'key': self.key} if self.key else {}
		response = await self.session.get(url, params=params)
		if response.status == 404:
			raise MXMapNotFound('Map has not been found!')
		if response.status == 302:
			raise MXInvalidResponse('Map author has declined info for the map. Status code: {}'.format(response.status))
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))
		record = list()
		for info in await response.json():
			record.append((info))
		return record
	
	async def map_offline_records(self, trackid):
		url = '{base}/replays/get_replays/{id}/10'.format(base=self.base_url(True), id=trackid)
		response = await self.session.get(url)
		if response.status == 404:
			raise MXMapNotFound('Map has not been found!')
		if response.status == 302:
			raise MXInvalidResponse('Map author has declined info for the map. Status code: {}'.format(response.status))
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))
		record = list()
		for info in await response.json():
			print(info)
			record.append((info))
		return record
	
	async def map_info_page(self, *ids):
		if self.site != 'sm':
			url = '{base}/maps/get_map_info/multi/{ids}'.format(
			base=self.base_url(True),
			ids=','.join(str(i) for i in ids[0])
			)
			
		else:
			url = '{base}/maps/{ids}'.format(
				base=self.base_url(True),
				ids=','.join(str(i) for i in ids[0])
			)

		params = {'key': self.key} if self.key else {}
		response = await self.session.get(url, params=params)
		if response.status == 404:
			raise MXMapNotFound('Map has not been found!')
		if response.status == 302:
			raise MXInvalidResponse('Map author has declined info for the map. Status code: {}'.format(response.status))
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))
		maps = list()
		for info in await response.json():
			# Parse some differences between the api game endpoints.
			mx_id = info['TrackID'] if 'TrackID' in info else info['MapID']
			info['MapID'] = mx_id
			info['MapUID'] = info['TrackUID'] if 'TrackUID' in info else info['MapUID']
			maps.append((mx_id, info))
		return maps

	async def pack_info(self, id, token):
		url = '{base}/api/mappack/get_info/{id}?token={token}&secret={token}'.format(
			base=self.base_url(),
			id=id,
			token=token
		)
		params = {'key': self.key} if self.key else {}
		response = await self.session.get(url, params=params)
		if response.status == 404:
			raise MXMapNotFound('Map has not been found!')
		if response.status == 302:
			raise MXInvalidResponse('Map author has declined info for the map. Status code: {}'.format(response.status))
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))

		return response.json()

	async def get_pack_ids(self, pack_id, token):
		url = '{base}/api/mappack/get_mappack_tracks/{id}?token={token}'.format(
			base=self.base_url(),
			id=pack_id,
			token=token
		)
		params = {'key': self.key} if self.key else {}
		response = await self.session.get(url, params=params)
		if response.status == 404:
			raise MXMapNotFound('Map pack not found!')
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))
		maps = list()
		if response.content_length > 0:
			for info in await response.json():
				# Parse some differences between the api game endpoints.
				mx_id = info['TrackID']
				maps.append((mx_id, info))

			return maps
		else:
			raise MXMapNotFound("Mx returned with empty response.")

	async def download(self, mx_id):
		url = '{base}/maps/download/{id}'.format(
			base=self.base_url(),
			id=mx_id,
		)
		params = {'key': self.key} if self.key else {}
		response = await self.session.get(url, params=params)
		if response.status == 404:
			raise MXMapNotFound('Map has not been found!')
		if response.status == 302:
			raise MXInvalidResponse(
				'Map author has declined download of the map. Status code: {}'.format(response.status))
		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))
		return response



import asyncio
import logging
import aiohttp

from . import semver


class _UpdateChecker:  # pragma: no cover

	def __init__(self):
		self.latest = None
		self.current = None
		self.instance = None
		self.releases = list()

		self.url = 'https://api.github.com/repos/{}/releases'.format('PyPlanet/PyPlanet')

	async def init_checker(self, instance):
		"""
		Initiate checker.

		:param instance: Instance of controller.
		:type instance: pyplanet.core.instance.Instance
		"""
		self.instance = instance
		self.instance.signals.listen('maniaplanet:player_connect', self.connect)

		asyncio.ensure_future(self.check(True))
		asyncio.ensure_future(self.loop())

	async def loop(self):
		while True:
			await asyncio.sleep(7200)  # 2 hours.
			try:
				await self.check()
			except:
				pass

	async def check(self, first_check=False):
		from pyplanet import __version__ as current_version
		running_prerelease = semver.is_prerelease(current_version)

		logging.debug('Checking for new versions...')

		async with aiohttp.ClientSession() as session:
			async with session.get(self.url) as resp:
				self.releases = list()
				for release in await resp.json():
					self.releases.append(release['tag_name'])

				for release in await resp.json():
					if release['draft'] or (not running_prerelease and release['prerelease']):
						continue

					self.latest = release['tag_name']
					break

				self.current = current_version
				logging.debug('Version check, your version: {}, online version: {}'.format(self.current, self.latest))

				if first_check and self.update_available:
					logging.info('New version of PyPlanet available, consider updating: {}'.format(self.latest))
					await self.instance.chat(
						'\uf1e6 $FD4$oPy$369Planet$z$s$fff \uf0e7 new version available: v{}. Consider updating!'.format(self.latest)
					)

	@property
	def update_available(self):
		if self.latest is None or self.current is None:
			return False
		return semver.compare(self.latest, self.current) > 0

	async def connect(self, player, **kwargs):
		if player.level > 0 and self.update_available is not False:
			await self.instance.gbx.multicall(
				self.instance.chat(
					'\uf1e6 $FD4$oPy$369Planet$z$s$fff \uf0e7 new version available: v{}. Consider updating!'.format(self.latest),
					player
				),
				self.instance.chat(
					'$fffClick $l[http://pypla.net/en/stable/intro/upgrading.html]here$l to open the upgrade instructions.',
					player
				)
			)


UpdateChecker = _UpdateChecker()
